# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import copy
import os
import time

from mmcv import Config, ConfigDict, build_from_cfg

from .modules.hooks.workflow_hooks import WorkflowHook, build_workflow_hook
from .registry import STAGES
from .stage import get_available_types
from .utils.config_utils import MPAConfig
from .utils.logger import config_logger, get_logger
from .workflow import Workflow

# from collections import defaultdict


logger = get_logger()


def __build_stage(config, common_cfg=None, index=0):
    logger.info("build_stage()")
    logger.debug(f"[args] config = {config}, common_cfg = {common_cfg}, index={index}")
    config.type = config.type if "type" in config.keys() else "Stage"  # TODO: tmp workaround code for competability
    config.common_cfg = common_cfg
    config.index = index
    return build_from_cfg(config, STAGES)


def __build_workflow(config):
    logger.info("build_workflow()")
    logger.debug(f"[args] config = {config}")

    whooks = []
    whooks_cfg = config.get("workflow_hooks", [])
    for whook_cfg in whooks_cfg:
        if isinstance(whook_cfg, WorkflowHook):
            whooks.append(whook_cfg)
        else:
            whook = build_workflow_hook(whook_cfg.copy())
            whooks.append(whook)

    output_path = config.get("output_path", "logs")
    folder_name = f"{time.strftime('%Y%m%d_%H%M%S', time.localtime())}"
    config.output_path = os.path.join(output_path, folder_name)
    os.makedirs(config.output_path, exist_ok=True)

    # create symbolic link to the output path
    symlink_dst = os.path.join(output_path, "latest")
    if os.path.exists(symlink_dst):
        os.unlink(symlink_dst)
    os.symlink(folder_name, symlink_dst, True)

    log_level = config.get("log_level", "INFO")
    config_logger(os.path.join(config.output_path, "app.log"), level=log_level)

    if not hasattr(config, "gpu_ids"):
        gpu_ids = os.environ.get("CUDA_VISIBLE_DEVICES", None)
        logger.info(f"CUDA_VISIBLE_DEVICES = {gpu_ids}")
        if gpu_ids is not None:
            if isinstance(gpu_ids, str):
                config.gpu_ids = range(len(gpu_ids.split(",")))
            else:
                raise ValueError(f"not supported type for gpu_ids: {type(gpu_ids)}")
        else:
            config.gpu_ids = range(1)

    common_cfg = copy.deepcopy(config)
    common_cfg.pop("stages")
    if len(whooks_cfg) > 0:
        common_cfg.pop("workflow_hooks")

    stages = [__build_stage(stage_cfg.copy(), common_cfg, index=i) for i, stage_cfg in enumerate(config.stages)]
    return Workflow(stages, whooks)


def build(config, mode=None, stage_type=None, common_cfg=None):
    logger.info("called build_recipe()")
    logger.debug(f"[args] config = {config}")

    if not isinstance(config, Config):
        if isinstance(config, str):
            if os.path.exists(config):
                config = MPAConfig.fromfile(config)
            else:
                logger.error(f"cannot find configuration file {config}")
                raise ValueError(f"cannot find configuration file {config}")

    if hasattr(config, "stages"):
        # build as workflow
        return __build_workflow(config)
    else:
        # build as stage
        if not hasattr(config, "type"):
            logger.info("seems to be passed stage yaml...")
            supported_stage_types = get_available_types()
            if stage_type in supported_stage_types:
                cfg_dict = ConfigDict(
                    dict(
                        type=stage_type,
                        name=f"{stage_type}-{mode}",
                        mode=mode,
                        config=config,
                        index=0,
                    )
                )
            else:
                msg = f"type {stage_type} is not in {supported_stage_types}"
                logger.error(msg)
                raise RuntimeError(msg)
        return __build_stage(cfg_dict, common_cfg=common_cfg)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .cls_head import ClsHead
from .conv_head import ConvClsHead

# flake8: noqa
from .mmov_cls_head import MMOVClsHead



from copy import deepcopy
from typing import Any, Dict

import pytest
import torch
import torch.nn as nn

from otx.algorithms.classification.adapters.mmcls import BYOL
from tests.test_suite.e2e_test_system import e2e_pytest_unit


@e2e_pytest_unit
class TestBYOL:
    """Test BYOL."""

    @pytest.fixture(autouse=True)
    def setup(self, monkeypatch, mocker) -> None:
        class MockBackbone(nn.Module):
            def __init__(self):
                super().__init__()
                self.pipeline = nn.Sequential(nn.Conv2d(3, 1, (1, 1), bias=False), nn.Conv2d(1, 1, (1, 1), bias=False))

            def init_weights(self, pretrained=None):
                pass

            def forward(self, x):
                return self.pipeline(x)

        class MockNeck(nn.Sequential):
            def __init__(self):
                super().__init__(nn.Linear(2, 2, bias=False), nn.Linear(2, 2, bias=False))

            def init_weights(self, init_linear=None):
                pass

        class MockHead(nn.Sequential):
            def __init__(self):
                super().__init__(nn.Linear(2, 2, bias=False))

            def init_weights(self, init_linear=None):
                pass

            def forward(self, *args, **kwargs):
                return {"loss": torch.Tensor(1)}

        def build_mock_backbone(*args, **kwargs):
            return MockBackbone()

        def build_mock_neck(*args, **kwargs):
            return MockNeck()

        def build_mock_head(*args, **kwargs):
            return MockHead()

        monkeypatch.setattr(
            "otx.algorithms.classification.adapters.mmcls.models.classifiers.byol.build_backbone", build_mock_backbone
        )
        monkeypatch.setattr(
            "otx.algorithms.classification.adapters.mmcls.models.classifiers.byol.build_neck", build_mock_neck
        )
        monkeypatch.setattr(
            "otx.algorithms.classification.adapters.mmcls.models.classifiers.byol.build_head", build_mock_head
        )

        self.byol = BYOL(backbone={}, neck={}, head={})

    @e2e_pytest_unit
    def test_init_weights(self) -> None:
        """Test init_weights function."""
        for param_ol, param_tgt in zip(self.byol.online_backbone.parameters(), self.byol.target_backbone.parameters()):
            assert torch.all(param_ol == param_tgt)
            assert param_ol.requires_grad
            assert not param_tgt.requires_grad

        for param_ol, param_tgt in zip(
            self.byol.online_projector.parameters(), self.byol.target_projector.parameters()
        ):
            assert torch.all(param_ol == param_tgt)
            assert param_ol.requires_grad
            assert not param_tgt.requires_grad

    @e2e_pytest_unit
    def test_momentum_update(self) -> None:
        """Test _momentum_update function."""
        original_params = {"backbone": [], "projector": []}
        for param_tgt in self.byol.target_backbone.parameters():
            param_tgt.data *= 2.0
            original_params["backbone"].append(deepcopy(param_tgt))

        for param_tgt in self.byol.target_projector.parameters():
            param_tgt.data *= 2.0
            original_params["projector"].append(deepcopy(param_tgt))

        self.byol.momentum_update()

        for param_ol, param_tgt, orig_tgt in zip(
            self.byol.online_backbone.parameters(), self.byol.target_backbone.parameters(), original_params["backbone"]
        ):
            assert torch.all(
                param_tgt.data == orig_tgt * self.byol.momentum + param_ol.data * (1.0 - self.byol.momentum)
            )

        for param_ol, param_tgt, orig_tgt in zip(
            self.byol.online_projector.parameters(),
            self.byol.target_projector.parameters(),
            original_params["projector"],
        ):
            assert torch.all(
                param_tgt.data == orig_tgt * self.byol.momentum + param_ol.data * (1.0 - self.byol.momentum)
            )

    @e2e_pytest_unit
    def test_train_step(self) -> None:
        """Test train_step function wraps forward and _parse_losses."""
        img1 = torch.randn((1, 3, 2, 2))
        img2 = torch.randn((1, 3, 2, 2))

        outputs = self.byol.train_step(data=dict(img1=img1, img2=img2), optimizer=None)

        assert "loss" in outputs
        assert "log_vars" in outputs
        assert "num_samples" in outputs

    @e2e_pytest_unit
    @pytest.mark.parametrize(
        "orig_state_dict,prefix,expected",
        [
            ({"online_backbone.layer": 1}, "", {"layer": 1}),
            ({"backbone.layer": 1}, "", {"backbone.layer": 1}),
            ({"backbone.layer": 1}, "backbone.", {"backbone.layer": 1}),
        ],
    )
    def test_state_dict_hook(self, orig_state_dict: Dict[str, Any], prefix: str, expected: Dict[str, Any]) -> None:
        """Test state_dict_hook function."""
        new_state_dict = BYOL.state_dict_hook(module=self.byol, state_dict=orig_state_dict, prefix=prefix)

        assert new_state_dict == expected



import os

import pytest

from otx.mpa.seg.semisl.inferrer import SemiSLSegInferrer
from otx.mpa.utils.config_utils import MPAConfig
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.segmentation.test_helpers import DEFAULT_SEG_TEMPLATE_DIR

SEMISL_RECIPE_CONFIG_PATH = "otx/recipes/stages/segmentation/semisl.py"


class TestOTXSegStage:
    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        cfg = MPAConfig.fromfile(SEMISL_RECIPE_CONFIG_PATH)
        self.inferrer = SemiSLSegInferrer(name="", mode="train", config=cfg, common_cfg=None, index=0)
        self.model_cfg = MPAConfig.fromfile(os.path.join(DEFAULT_SEG_TEMPLATE_DIR, "semisl/model.py"))
        self.data_cfg = MPAConfig.fromfile(os.path.join(DEFAULT_SEG_TEMPLATE_DIR, "semisl/data_pipeline.py"))

    @e2e_pytest_unit
    def test_configure(self):
        updated_cfg = self.inferrer.configure(self.model_cfg, "", self.data_cfg)

        assert "orig_type" not in updated_cfg
        assert "unsup_weight" not in updated_cfg
        assert "semisl_start_iter" not in updated_cfg



"""Wrapper for Open Model Zoo for Anomaly tasks."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Union

import numpy as np
from openvino.model_zoo.model_api.models import SegmentationModel
from openvino.model_zoo.model_api.models.types import NumericalValue


class AnomalyBase(SegmentationModel):
    """Wrapper for anomaly tasks."""

    __model__ = "anomaly_base"

    @classmethod
    def parameters(cls):
        """Dictionary containing model parameters."""
        parameters = super().parameters()
        parameters["resize_type"].update_default_value("standard")
        parameters.update(
            {
                "image_threshold": NumericalValue(description="Image-level Threshold value to locate anomaly"),
                "pixel_threshold": NumericalValue(description="Pixel-level Threshold value to locate anomaly"),
                "min": NumericalValue(description="Threshold value to locate anomaly"),
                "max": NumericalValue(description="Threshold value to locate anomaly"),
                "threshold": NumericalValue(description="Threshold used to classify anomaly"),
            }
        )

        return parameters

    @staticmethod
    def _normalize(
        targets: Union[np.ndarray, np.float32],
        threshold: Union[np.ndarray, float],
        min_val: Union[np.ndarray, float],
        max_val: Union[np.ndarray, float],
    ) -> np.ndarray:
        """Apply min-max normalization and shift the values such that the threshold value is centered at 0.5."""
        normalized = ((targets - threshold) / (max_val - min_val)) + 0.5
        if isinstance(targets, (np.ndarray, np.float32)):
            normalized = np.minimum(normalized, 1)
            normalized = np.maximum(normalized, 0)
        else:
            raise ValueError(f"Targets must be either Tensor or Numpy array. Received {type(targets)}")
        return normalized



"""Utils for dynamically importing stuff."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.


import importlib
import inspect
import json
import os

# pylint: disable=protected-access

SUPPORTED_BACKBONE_BACKENDS = {
    "otx": "otx.algorithms.common.adapters.mmcv.models",
    "mmcls": "mmcls.models",
    "mmdet": "mmdet.models",
    "mmseg": "mmseg.models",
    "torchvision": "otx.algorithms.common.adapters.mmcv.models",
    "pytorchcv": "mmdet.models",
    "omz.mmcls": "otx.mpa.modules.ov.models.mmcls.backbones.mmov_backbone",
}


def get_impl_class(impl_path):
    """Returns a class by its path in package."""

    task_impl_module_name, task_impl_class_name = impl_path.rsplit(".", 1)
    task_impl_module = importlib.import_module(task_impl_module_name)
    task_impl_class = getattr(task_impl_module, task_impl_class_name)

    return task_impl_class


def get_backbone_list(backend):
    """Gather available backbone list from json file & imported lib."""
    available_backbone_path = os.path.join(get_otx_root_path(), f"cli/builder/supported_backbone/{backend}.json")
    available_backbones = {}
    if os.path.exists(available_backbone_path):
        with open(available_backbone_path, "r", encoding="UTF-8") as f:
            available_backbones = json.load(f)
        available_backbones = available_backbones["backbones"]
    elif backend == "pytorchcv" and importlib.util.find_spec(backend):
        backbone_list = importlib.import_module(f"{backend}.model_provider")._models
        backbone_format = {"required": [], "options": {}, "available": []}
        for backbone in backbone_list:
            backbone_type = f"mmdet.{backbone}"
            available_backbones[backbone_type] = backbone_format
    else:
        raise ValueError(f"{backend} cannot be imported or supported.")
    return available_backbones


def get_backbone_registry(backend=None):
    """Gather backbone list from backends."""
    if backend not in SUPPORTED_BACKBONE_BACKENDS:
        raise ValueError(f"{backend} is an unsupported backbone backend.")

    custom_imports = []
    backend_import_path = SUPPORTED_BACKBONE_BACKENDS[backend]
    mm_backbones = importlib.import_module(backend_import_path)
    mm_registry = mm_backbones.BACKBONES
    custom_imports.append(backend_import_path)
    return mm_registry, custom_imports


def get_module_args(module):
    """Gather module's Required Args."""
    if module is None:
        return []
    required_args = []
    default_args = {}
    args_signature = inspect.signature(module)
    for arg_key, arg_value in args_signature.parameters.items():
        if arg_value.default is inspect.Parameter.empty:
            required_args.append(arg_key)
            continue
        default_args[arg_key] = arg_value.default
    # Get args from parents
    parent_module = module.__bases__
    while len(parent_module):
        parent_args_signature = inspect.signature(parent_module[0])
        for arg_key, arg_value in parent_args_signature.parameters.items():
            if arg_key == "depth" and "arch" in required_args:
                continue
            if arg_value.default is inspect.Parameter.empty and arg_key not in required_args:
                required_args.append(arg_key)
                continue
        parent_module = parent_module[0].__bases__
    required_args = [arg for arg in required_args if arg not in ("args", "kwargs", "self")]
    return required_args, default_args


def get_otx_root_path():
    """Get otx root path from importing otx."""
    otx_module = importlib.import_module("otx")
    if otx_module:
        return os.path.dirname(inspect.getfile(otx_module))
    return None



"""
General fixtures.
"""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import pytest

from tests.test_helpers import LabelSchemaExample


@pytest.fixture(scope="session")
def label_schema_example():
    """
    Returns a label schema example.
    """

    return LabelSchemaExample()



"""NNCF utils for mmdet."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# flake8: noqa

from .builder import build_nncf_detector

__all__ = [
    "build_nncf_detector",
]



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from copy import deepcopy
from typing import Dict, List, Optional, Union

import openvino.runtime as ov
from mmseg.models.builder import HEADS
from mmseg.models.decode_heads.decode_head import BaseDecodeHead

from ...mmov_model import MMOVModel


@HEADS.register_module()
class MMOVDecodeHead(BaseDecodeHead):
    def __init__(
        self,
        model_path_or_model: Union[str, ov.Model] = None,
        weight_path: Optional[str] = None,
        inputs: Dict[str, Union[str, List[str]]] = {},
        outputs: Dict[str, Union[str, List[str]]] = {},
        init_weight: bool = False,
        verify_shape: bool = True,
        *args,
        **kwargs,
    ):
        self._model_path_or_model = model_path_or_model
        self._weight_path = weight_path
        self._inputs = deepcopy(inputs)
        self._outputs = deepcopy(outputs)
        self._init_weight = init_weight
        self._verify_shape = verify_shape

        # dummy input
        channels = 1
        super().__init__(
            channels=channels,
            *args,
            **kwargs,
        )
        delattr(self, "channels")

        if "extractor" in inputs and "extractor" in outputs:
            self.extractor = MMOVModel(
                self._model_path_or_model,
                self._weight_path,
                inputs=inputs["extractor"],
                outputs=outputs["extractor"],
                remove_normalize=False,
                merge_bn=True,
                paired_bn=True,
                verify_shape=self._verify_shape,
                init_weight=self._init_weight,
            )

        assert "cls_seg" in inputs and "cls_seg" in outputs
        self.conv_seg = MMOVModel(
            self._model_path_or_model,
            self._weight_path,
            inputs=inputs["cls_seg"],
            outputs=outputs["cls_seg"],
            remove_normalize=False,
            merge_bn=False,
            paired_bn=False,
            verify_shape=self._verify_shape,
            init_weight=self._init_weight,
        )

    def init_weights(self):
        # TODO
        pass

    def forward(self, inputs):
        outputs = self._transform_inputs(inputs)
        if getattr(self, "extractor"):
            outputs = self.extractor(outputs)
        outputs = self.cls_seg(outputs)
        return outputs



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Dict, List

from mmcls.models.builder import BACKBONES

from ....graph.parsers.cls.cls_base_parser import cls_base_parser
from ...mmov_model import MMOVModel


@BACKBONES.register_module()
class MMOVBackbone(MMOVModel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def parser(graph, **kwargs) -> Dict[str, List[str]]:
        output = cls_base_parser(graph, "backbone")
        if output is None:
            raise ValueError("Parser can not determine input and output of model. " "Please provide them explicitly")
        return output

    def init_weights(self, pretrained=None):
        # TODO
        pass



"""This file defines the URL class."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os
from typing import Union


class URL:
    """Class for working with URLs.

    Provides functions for getting components of the URL such as <scheme>://<host>/<path>
    """

    def __parse(self, url):
        """Parse URL into 3 internal saved components.

        (Function stolen from urllib library)
        <scheme>://<netloc>/<path>
        """

        def __splitnetloc(url, start=0):
            delimiter = len(url)  # position of end of domain part of url, default is end
            for char in "/?#":  # look for delimiters; the order is NOT important
                position = url.find(char, start)  # find first of this delimiter
                if position >= 0:  # if found
                    delimiter = min(delimiter, position)  # use earliest delimiter position
            return url[start:delimiter], url[delimiter:]  # return (domain, rest)

        if url is None:
            raise ValueError("Cannot build an URL from None")

        scheme_chars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789+-."
        netloc = scheme = ""
        i = url.find(":")

        if i > 0:
            for character in url[:i]:
                if character not in scheme_chars:
                    break
            else:
                rest = url[i + 1 :]  # make sure "url" is not actually a port number (in which case
                if not rest or any(c not in "0123456789" for c in rest):
                    scheme, url = url[:i].lower(), rest  # Not a port

        if url[:2] == "//":
            netloc, url = __splitnetloc(url, 2)
            if ("[" in netloc and "]" not in netloc) or ("]" in netloc and "[" not in netloc):
                raise ValueError("Invalid IPv6 URL")

        if "#" in url:
            url, _ = url.split("#", 1)
        if "?" in url:
            url, _ = url.split("?", 1)

        self.__scheme = scheme
        self.__netloc = netloc
        self.__path = url

    def __init__(self, url: Union["URL", str]):
        self.__url = str(url)
        self.__scheme = ""
        self.__netloc = ""
        self.__path = ""
        self.__parse(self.__url)

    def __repr__(self):
        """Returns representation of url."""
        return f"URL({self.__url})"

    @property
    def extension(self):
        """Gets the extension of the URL."""
        basename = os.path.basename(self.__path)
        return basename.rsplit(".")[-1] if "." in basename else ""

    @property
    def scheme(self):
        """Gets the scheme of the URL."""
        return self.__scheme

    @property
    def host(self):
        """Gets the host of the URL."""
        return self.__netloc

    @property
    def path(self):
        """Returns the full path to the full."""
        return self.__path

    @property
    def relative_path(self):
        """Returns the relative path to the full."""
        path = self.path
        while path.startswith("/"):
            path = path[1:]
        return path

    def __str__(self):
        """Returns url as string depending on the value of scheme and host."""
        # In this case, it is allowed to violate the guidelines, and to use multiple return statements.
        # As it improves readability of the code
        if self.scheme == "":
            return f"{self.__path}"
        if self.host == "":
            return f"{self.scheme}:{self.__path}"
        return f"{self.scheme}://{self.host}{self.__path}"

    def __hash__(self):
        """Returns hash of url."""
        return hash(str(self))

    def __eq__(self, other):
        """Checks if the url strings are equal."""
        return str(self) == str(other)

    def __len__(self):
        """Returns length of url."""
        return len(self.__url)



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .inferrer import IncrClsInferrer
from .trainer import IncrClsTrainer

__all__ = ["IncrClsTrainer", "IncrClsInferrer"]



"""Evaluate mean AP for segmentation."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from multiprocessing import Pool
from typing import Dict, List

import mmcv
import numpy as np
import pycocotools.mask as mask_util
from mmcv.utils import print_log
from mmdet.core.evaluation.class_names import get_classes
from mmdet.core.evaluation.mean_ap import average_precision
from mmdet.core.mask.structures import PolygonMasks
from terminaltables import AsciiTable


def print_map_summary(  # pylint: disable=too-many-locals,too-many-branches
    mean_ap, results, dataset=None, scale_ranges=None, logger=None
):
    """Print mAP/mIoU and results of each class.

    A table will be printed to show the gts/dets/recall/AP/IoU of each class
    and the mAP/mIoU.

    Args:
        mean_ap (float): Calculated from `eval_map()`.
        results (list[dict]): Calculated from `eval_map()`.
        dataset (list[str] | str | None): Dataset name or dataset classes.
        scale_ranges (list[tuple] | None): Range of scales to be evaluated.
        logger (logging.Logger | str | None): The way to print the mAP
            summary. See `mmcv.utils.print_log()` for details. Default: None.
    """

    if logger == "silent":
        return

    if isinstance(results[0]["ap"], np.ndarray):
        num_scales = len(results[0]["ap"])
    else:
        num_scales = 1

    if scale_ranges is not None:
        assert len(scale_ranges) == num_scales

    num_classes = len(results)

    recalls = np.zeros((num_scales, num_classes), dtype=np.float32)
    aps = np.zeros((num_scales, num_classes), dtype=np.float32)
    num_gts = np.zeros((num_scales, num_classes), dtype=int)
    mious = np.zeros((num_scales, num_classes), dtype=np.float32)
    for i, cls_result in enumerate(results):
        if cls_result["recall"].size > 0:
            recalls[:, i] = np.array(cls_result["recall"], ndmin=2)[:, -1]
        aps[:, i] = cls_result["ap"]
        mious[:, i] = cls_result["miou"]
        num_gts[:, i] = cls_result["num_gts"]

    if dataset is None:
        label_names = [str(i) for i in range(num_classes)]
    elif mmcv.is_str(dataset):
        label_names = get_classes(dataset)
    else:
        label_names = dataset

    if not isinstance(mean_ap, list):
        mean_ap = [mean_ap]

    header = ["class", "gts", "dets", "recall", "ap", "miou"]
    for i in range(num_scales):
        if scale_ranges is not None:
            print_log(f"Scale range {scale_ranges[i]}", logger=logger)
        table_data = [header]
        for j in range(num_classes):
            row_data = [
                label_names[j],
                num_gts[i, j],
                results[j]["num_dets"],
                f"{recalls[i, j]:.3f}",
                f"{aps[i, j]:.3f}",
                f"{mious[i, j]:.3f}",
            ]
            table_data.append(row_data)
        table_data.append(["mAP", "", "", "", f"{mean_ap[i]:.3f}", f"{np.mean(mious[i]):.3f}"])
        table = AsciiTable(table_data)
        table.inner_footing_row_border = True
        print_log("\n" + table.table, logger=logger)


def tpfpmiou_func(  # pylint: disable=too-many-locals
    det_masks: List[Dict],
    gt_masks: List[Dict],
    cls_scores,
    iou_thr=0.5,
):
    """Calculate Mean Intersection and Union (mIoU) and AP across predicted masks and GT masks.

    Args:
        det_masks: Detected masks of this image, with list size (m)
        gt_masks: GT masks of this image, with list size (n).
        cls_scores: (n, 1)
        iou_thr (float): IoU threshold to be considered as matched.
            Default: 0.5.

    Returns:
        tuple[np.ndarray, np.ndarray, float]:
            <tp, fp> with elements of 0 and 1, the shape of each array is (m).
            <gt_covered_iou>: the average IoU between predicted mask and GT mask
    """
    num_dets = len(det_masks)
    num_gts = len(gt_masks)

    tp = np.zeros(num_dets, dtype=np.float32)  # pylint: disable=invalid-name
    fp = np.zeros(num_dets, dtype=np.float32)  # pylint: disable=invalid-name
    gt_covered_iou = np.zeros(num_gts, dtype=np.float32)

    if len(gt_masks) == 0:
        fp[...] = 1
        return tp, fp, 0.0
    if num_dets == 0:
        return tp, fp, 0.0

    ious = mask_util.iou(det_masks, gt_masks, len(gt_masks) * [0])
    # for each det, the max iou with all gts
    ious_max = ious.max(axis=1)
    # for each det, which gt overlaps most with it
    ious_argmax = ious.argmax(axis=1)
    # sort all dets in descending order by scores
    sort_inds = np.argsort(-cls_scores)

    gt_covered = np.zeros(num_gts, dtype=bool)
    # if no area range is specified, gt_area_ignore is all False
    for i in sort_inds:
        if ious_max[i] >= iou_thr:
            matched_gt = ious_argmax[i]
            if not gt_covered[matched_gt]:
                gt_covered[matched_gt] = True
                gt_covered_iou[matched_gt] = ious_max[i]
                tp[i] = 1
            else:
                fp[i] = 1
            # otherwise ignore this detected bbox, tp = 0, fp = 0
        else:
            fp[i] = 1
    return tp, fp, np.mean(gt_covered_iou)


def get_cls_results(det_results, annotations, class_id):
    """Get det results and gt information of a certain class.

    Args:
        det_results (list[tuple]): Same as `eval_segm()`.
        annotations (list[dict]): Same as `eval_segm()`.
        class_id (int): ID of a specific class.

    Returns:
        tuple[list[list[dict]], list[list[dict]], list[np.ndarray]]:
          <detected masks>: list[dict] includes all the predicted RLE masks
              for an image.
          <gt masks>: list[dict] includes all the GT RLE masks for an image
          <detected box scores>: each array in the list describes the predicted
              box scores for an image.
    """
    cls_scores = [img_res[0][class_id][..., -1] for img_res in det_results]
    cls_dets = []
    for i, det in enumerate(det_results):
        det_masks = det[1][class_id]
        cls_dets.append([])
        for det_mask in det_masks:
            if isinstance(det_mask, np.ndarray):
                cls_dets[i].append(mask_util.encode(np.array(det_mask[:, :, np.newaxis], order="F", dtype="uint8"))[0])
            else:
                cls_dets[i].append(det_mask)

    cls_gts = []
    for ann in annotations:
        gt_inds = ann["labels"] == class_id
        if isinstance(ann["masks"], PolygonMasks):
            masks = ann["masks"].to_ndarray()[gt_inds]
            encoded_masks = [
                mask_util.encode(np.array(m[:, :, np.newaxis], order="F", dtype="uint8"))[0] for m in masks
            ]
            cls_gts.append(encoded_masks)
        elif isinstance(ann["masks"], list):
            cls_gts.append([])
        else:
            raise RuntimeError("Unknown annotation format")

    return cls_dets, cls_gts, cls_scores


def eval_segm(  # pylint: disable=too-many-locals
    det_results,
    annotations,
    iou_thr=0.5,
    dataset=None,
    logger=None,
    nproc=4,
    metric="mAP",
):
    """Evaluate mAP/mIoU of a dataset.

    Args:
        det_results (list[tuple[list, list]]): [[cls1_det, cls2_det, ...], ...].
            The outer list indicates images, and the inner tuple indicates box
            and mask prediction. Each list of a predicted type (box/mask)
            includes the per-class detection of that type.
        annotations (list[dict]): Ground truth annotations where each item of
            the list indicates an image. Keys of annotations are:
            - `bboxes`: numpy array of shape (n, 4)
            - `labels`: numpy array of shape (n, )
            - `masks`: numpy array of shape (k, 4)
        iou_thr (float): IoU threshold to be considered as matched.
            Default: 0.5.
        dataset (list[str] | str | None): Dataset name or dataset classes,
            there are minor differences in metrics for different datsets, e.g.
            "voc07", "imagenet_det", etc. Default: None.
        logger (logging.Logger | str | None): The way to print the mAP
            summary. See `mmcv.utils.print_log()` for details. Default: None.
        nproc (int): Processes used for computing tpfpmiou_func. Default: 4.

    Returns:
        tuple: (mIoU, [dict, dict, ...])
    """
    assert len(det_results) == len(annotations)

    num_imgs = len(det_results)
    num_classes = len(det_results[0][0])

    pool = Pool(nproc)  # pylint: disable=consider-using-with
    eval_results = []
    for i in range(num_classes):
        # get gt and det bboxes of this class
        cls_results = get_cls_results(det_results, annotations, i)
        cls_dets, cls_gts, cls_scores = cls_results

        # compute tp and fp for each image with multiple processes
        tpfpmiou = pool.starmap(tpfpmiou_func, zip(cls_dets, cls_gts, cls_scores, [iou_thr for _ in range(num_imgs)]))
        tp, fp, miou = tuple(zip(*tpfpmiou))  # pylint: disable=invalid-name

        # sort all det bboxes by score, also sort tp and fp
        cls_scores = np.hstack(cls_scores)
        num_dets = cls_scores.shape[0]
        num_gts = np.sum([len(cls_gts) for cls_gts in cls_gts])
        sort_inds = np.argsort(cls_scores)[::-1]
        tp = np.hstack(tp)[sort_inds]  # pylint: disable=invalid-name
        fp = np.hstack(fp)[sort_inds]  # pylint: disable=invalid-name
        # calculate recall and precision with tp and fp
        tp = np.cumsum(tp)  # pylint: disable=invalid-name
        fp = np.cumsum(fp)  # pylint: disable=invalid-name
        eps = np.finfo(np.float32).eps
        recalls = tp / np.maximum(num_gts, eps)
        precisions = tp / np.maximum((tp + fp), eps)
        miou = np.mean(np.stack(miou))
        # calculate AP
        mode = "area" if dataset != "voc07" else "11points"
        ap = average_precision(recalls, precisions, mode)  # pylint: disable=invalid-name
        eval_results.append(
            {
                "num_gts": num_gts,
                "num_dets": num_dets,
                "recall": recalls,
                "precision": precisions,
                "ap": ap,
                "miou": miou,
            }
        )
    pool.close()

    metrics = {"mAP": 0.0, "mIoU": 0.0}
    mious, aps = [], []
    for cls_result in eval_results:
        if cls_result["num_gts"] > 0:
            aps.append(cls_result["ap"])
            mious.append(cls_result["miou"])
    mean_ap = np.array(aps).mean().item() if aps else 0.0
    mean_miou = np.array(mious).mean().item() if mious else 0.0
    metrics["mAP"] = mean_ap
    metrics["mIoU"] = mean_miou

    print_map_summary(mean_ap, eval_results, dataset, None, logger=logger)

    return metrics[metric], eval_results



"""Utils for working with Configurable parameters."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from pathlib import Path

import yaml


def override_parameters(overrides, parameters):
    """Overrides parameters values by overrides."""

    allowed_keys = {"default_value", "value"}
    for k, val in overrides.items():
        if isinstance(val, dict):
            if k in parameters.keys():
                override_parameters(val, parameters[k])
            else:
                raise ValueError(f'The "{k}" is not in original parameters.')
        else:
            if k in allowed_keys:
                parameters[k] = val
            else:
                raise ValueError(f'The "{k}" is not in allowed_keys: {allowed_keys}')


def configure_dataset(args, data_yaml_path=None):
    """Configure dataset args."""

    # Create instances of Task, ConfigurableParameters and Dataset.
    data_subset_format = {"ann-files": None, "data-roots": None}
    data_config = {"data": {subset: data_subset_format.copy() for subset in ("train", "val", "test")}}
    data_config["data"]["unlabeled"] = {"file-list": None, "data-roots": None}
    if data_yaml_path and Path(data_yaml_path).exists():
        with open(Path(data_yaml_path), "r", encoding="UTF-8") as stream:
            data_config = yaml.safe_load(stream)

    # The command's args are overridden and use first
    if "train_ann_files" in args and args.train_ann_files:
        data_config["data"]["train"]["ann-files"] = str(Path(args.train_ann_files).absolute())
    if "train_data_roots" in args and args.train_data_roots:
        data_config["data"]["train"]["data-roots"] = str(Path(args.train_data_roots).absolute())
    if "val_ann_files" in args and args.val_ann_files:
        data_config["data"]["val"]["ann-files"] = str(Path(args.val_ann_files).absolute())
    if "val_data_roots" in args and args.val_data_roots:
        data_config["data"]["val"]["data-roots"] = str(Path(args.val_data_roots).absolute())
    if "unlabeled_file_list" in args and args.unlabeled_file_list:
        data_config["data"]["unlabeled"]["file-list"] = str(Path(args.unlabeled_file_list).absolute())
    if "unlabeled_data_roots" in args and args.unlabeled_data_roots:
        data_config["data"]["unlabeled"]["data-roots"] = str(Path(args.unlabeled_data_roots).absolute())
    if "test_ann_files" in args and args.test_ann_files:
        data_config["data"]["test"]["ann-files"] = str(Path(args.test_ann_files).absolute())
    if "test_data_roots" in args and args.test_data_roots:
        data_config["data"]["test"]["data-roots"] = str(Path(args.test_data_roots).absolute())
    return data_config



"""Inference Callbacks for OTX inference."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import Any, List

import numpy as np
import pytorch_lightning as pl
import torch
from anomalib.models import AnomalyModule
from anomalib.post_processing import anomaly_map_to_color_map
from pytorch_lightning.callbacks import Callback
from torch import Tensor

from otx.algorithms.anomaly.adapters.anomalib.logger import get_logger
from otx.api.entities.annotation import Annotation
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label import LabelEntity
from otx.api.entities.model_template import TaskType
from otx.api.entities.result_media import ResultMediaEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.utils.segmentation_utils import create_annotation_from_segmentation_map

logger = get_logger(__name__)


class AnomalyInferenceCallback(Callback):
    """Callback that updates the OTX dataset during inference."""

    def __init__(self, otx_dataset: DatasetEntity, labels: List[LabelEntity], task_type: TaskType):
        self.otx_dataset = otx_dataset
        self.normal_label = [label for label in labels if not label.is_anomalous][0]
        self.anomalous_label = [label for label in labels if label.is_anomalous][0]
        self.task_type = task_type
        self.label_map = {0: self.normal_label, 1: self.anomalous_label}

    def on_predict_epoch_end(self, _trainer: pl.Trainer, _pl_module: AnomalyModule, outputs: List[Any]):
        """Call when the predict epoch ends."""
        # TODO; refactor Ignore too many locals
        # pylint: disable=too-many-locals
        outputs = outputs[0]
        # collect generic predictions
        pred_scores = torch.hstack([output["pred_scores"].cpu() for output in outputs])
        pred_labels = torch.hstack([output["pred_labels"].cpu() for output in outputs])
        anomaly_maps = torch.vstack([output["anomaly_maps"].cpu() for output in outputs])
        pred_masks = torch.vstack([output["pred_masks"].cpu() for output in outputs])

        # add the predictions to the dataset item depending on the task type
        if self.task_type == TaskType.ANOMALY_CLASSIFICATION:
            self._process_classification_predictions(pred_labels, pred_scores)
        elif self.task_type == TaskType.ANOMALY_DETECTION:
            # collect detection predictions
            pred_boxes = []
            box_scores = []
            box_labels = []
            for output in outputs:
                pred_boxes.extend(output["pred_boxes"])
                box_scores.extend(output["box_scores"])
                box_labels.extend(output["box_labels"])

            self._process_detection_predictions(pred_boxes, box_scores, box_labels, pred_scores, pred_masks.shape[-2:])
        elif self.task_type == TaskType.ANOMALY_SEGMENTATION:
            self._process_segmentation_predictions(pred_masks, anomaly_maps, pred_scores)

        # add anomaly map as metadata
        for dataset_item, anomaly_map in zip(self.otx_dataset, anomaly_maps):
            dataset_item.append_metadata_item(
                ResultMediaEntity(
                    name="Anomaly Map",
                    type="anomaly_map",
                    annotation_scene=dataset_item.annotation_scene,
                    numpy=anomaly_map_to_color_map(anomaly_map.squeeze().numpy(), normalize=False),
                )
            )

    def _process_classification_predictions(self, pred_labels: Tensor, pred_scores: Tensor):
        """Add classification predictions to the dataset items.

        Args:
            pred_labels (Tensor): Predicted image labels.
            pred_scores (Tensor): Predicted image-level anomaly scores.
        """
        for dataset_item, pred_label, pred_score in zip(self.otx_dataset, pred_labels, pred_scores):
            # get label
            label = self.anomalous_label if pred_label else self.normal_label
            probability = pred_score if pred_label else 1 - pred_score
            # update dataset item
            dataset_item.append_labels([ScoredLabel(label=label, probability=float(probability))])

    def _process_detection_predictions(
        self,
        pred_boxes: List[Tensor],
        box_scores: List[Tensor],
        box_labels: List[Tensor],
        pred_scores: Tensor,
        image_size: torch.Size,
    ):
        """Add detection predictions to the dataset items.

        Args:
            pred_boxes (List[Tensor]): Predicted bounding box locations.
            box_scores (List[Tensor]): Predicted anomaly scores for the bounding boxes.
            box_labels (List[Tensor]): Predicted labels for the bounding boxes.
            pred_scores (Tensor): Predicted image-level anomaly scores.
            image_size: (torch.Size): Image size of the original images.
        """
        # TODO; refactor Ignore too many locals
        # pylint: disable=too-many-locals
        height, width = image_size
        for dataset_item, im_boxes, im_box_scores, im_box_labels, pred_score in zip(
            self.otx_dataset, pred_boxes, box_scores, box_labels, pred_scores
        ):
            # generate annotations
            annotations: List[Annotation] = []
            for box, score, label in zip(im_boxes, im_box_scores, im_box_labels):
                if box[0] >= box[2] or box[1] >= box[3]:  # discard 1-pixel boxes
                    continue
                shape = Rectangle(
                    x1=box[0].item() / width,
                    y1=box[1].item() / height,
                    x2=box[2].item() / width,
                    y2=box[3].item() / height,
                )
                label = self.label_map[label.item()]
                probability = score.item()
                annotations.append(Annotation(shape=shape, labels=[ScoredLabel(label=label, probability=probability)]))
            # get label
            label = self.anomalous_label if annotations else self.normal_label
            probability = pred_score if label.is_anomalous else 1 - pred_score
            # update dataset item
            dataset_item.append_annotations(annotations)
            dataset_item.append_labels([ScoredLabel(label=label, probability=float(probability))])

    def _process_segmentation_predictions(self, pred_masks: Tensor, anomaly_maps: Tensor, pred_scores: Tensor):
        """Add segmentation predictions to the dataset items.

        Args:
            pred_masks (Tensor): Predicted anomaly masks.
            anomaly_maps (Tensor): Predicted pixel-level anomaly scores.
            pred_scores (Tensor): Predicted image-level anomaly scores.
        """
        for dataset_item, pred_mask, anomaly_map, pred_score in zip(
            self.otx_dataset, pred_masks, anomaly_maps, pred_scores
        ):
            # generate polygon annotations
            annotations = create_annotation_from_segmentation_map(
                hard_prediction=pred_mask.squeeze().numpy().astype(np.uint8),
                soft_prediction=anomaly_map.squeeze().numpy(),
                label_map=self.label_map,
            )
            # get label
            label = self.normal_label if len(annotations) == 0 else self.anomalous_label
            probability = pred_score if label.is_anomalous else 1 - pred_score
            # update dataset item
            dataset_item.append_annotations(annotations)
            dataset_item.append_labels([ScoredLabel(label=label, probability=float(probability))])



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from mmcv.runner import HOOKS, Hook


@HOOKS.register_module()
class SaveInitialWeightHook(Hook):
    def __init__(self, save_path, file_name: str = "weights.pth", **kwargs):
        self._save_path = save_path
        self._file_name = file_name
        self._args = kwargs

    def before_run(self, runner):
        runner.logger.info("Saving weight before training")
        runner.save_checkpoint(
            self._save_path, filename_tmpl=self._file_name, save_optimizer=False, create_symlink=False, **self._args
        )



__img_norm_cfg = dict(mean=[123.675, 116.28, 103.53], std=[58.395, 57.12, 57.375], to_rgb=True)

train_pipeline = [
    dict(type="LoadImageFromFile"),
    dict(type="LoadAnnotations", with_bbox=True),
    dict(type="Resize", img_scale=(1333, 800), keep_ratio=True),
    dict(type="RandomFlip", flip_ratio=0.5),
    dict(type="Normalize", **__img_norm_cfg),
    dict(type="Pad", size_divisor=32),
    dict(type="DefaultFormatBundle"),
    dict(type="Collect", keys=["img", "gt_bboxes", "gt_labels"]),
]

test_pipeline = [
    dict(type="LoadImageFromFile"),
    dict(
        type="MultiScaleFlipAug",
        img_scale=(1333, 800),
        flip=False,
        transforms=[
            dict(type="Resize", keep_ratio=True),
            dict(type="RandomFlip"),
            dict(type="Normalize", **__img_norm_cfg),
            dict(type="Pad", size_divisor=32),
            dict(type="ImageToTensor", keys=["img"]),
            dict(type="Collect", keys=["img"]),
        ],
    ),
]



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from mmcls.models.builder import HEADS
from mmcls.models.heads import ClsHead as OriginClsHead


@HEADS.register_module(force=True)
class ClsHead(OriginClsHead):
    def __init__(self, *args, **kwargs):
        do_squeeze = kwargs.pop("do_squeeze", False)
        super(ClsHead, self).__init__(*args, **kwargs)
        self._do_squeeze = do_squeeze

    def forward_train(self, cls_score, gt_label):
        if self._do_squeeze:
            cls_score = cls_score.unsqueeze(0).squeeze()
        return super().forward_train(cls_score, gt_label)

    def simple_test(self, cls_score):
        if self._do_squeeze:
            cls_score = cls_score.unsqueeze(0).squeeze()
        return super().simple_test(cls_score)



"""This module contains the definition for the `create` function within the configuration helper.

This function can be used to create a OTX configuration object from a dictionary or yaml representation.
"""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


from __future__ import annotations

import copy
from enum import Enum
from typing import Dict, List, TypeVar, Union

import attr
from omegaconf import DictConfig, OmegaConf

from otx.api.configuration.configurable_parameters import ConfigurableParameters
from otx.api.configuration.elements import (
    ConfigurableEnum,
    ParameterGroup,
    metadata_keys,
)
from otx.api.configuration.enums import AutoHPOState, ModelLifecycle
from otx.api.configuration.enums.config_element_type import (
    ConfigElementType,
    ElementCategory,
)
from otx.api.configuration.enums.utils import get_enum_names
from otx.api.configuration.ui_rules.rules import NullUIRules, Rule, UIRules

from .config_element_mapping import (
    GroupElementMapping,
    PrimitiveElementMapping,
    RuleElementMapping,
)
from .utils import deserialize_enum_value, input_to_config_dict

ParameterGroupTypeVar = TypeVar("ParameterGroupTypeVar", bound=ParameterGroup)
ExposureTypeVar = TypeVar("ExposureTypeVar", UIRules, Rule)

METADATA_ENUMS = {
    metadata_keys.AFFECTS_OUTCOME_OF: ModelLifecycle,
    metadata_keys.AUTO_HPO_STATE: AutoHPOState,
}


def construct_attrib_from_dict(dict_object: Union[dict, DictConfig]) -> ExposureTypeVar:
    """Constructs a ui exposure element from an input dictionary or DictConfig.

    Elements are mapped according to the 'type' field in the input dict.

    Args:
        dict_object (Union[dict, DictConfig]): Dictionary containing the arguments for the element constructor.

    Returns:
        ExposureTypeVar: Rule or UIRules element, constructed according to the input dict_object.
    """
    value_dict = dict(dict_object)
    object_type = str(value_dict.pop(metadata_keys.TYPE))
    if object_type in get_enum_names(RuleElementMapping):
        mapping = RuleElementMapping
    else:
        raise ValueError(f"Invalid type found in configuration dictionary: {object_type}")
    cls_constructor = mapping[object_type].value
    return cls_constructor(**value_dict)


def construct_ui_rules_from_dict(ui_exposure_settings: Union[dict, DictConfig]) -> UIRules:
    """Takes a dictionary representation of ui exposure logic and constructs an UIRules element out of this.

    Args:
        ui_exposure_settings: dictionary representing the logic to govern exposure of a parameter in the UI

    Returns:
        Exposure element constructed according to the settings passed in ui_exposure_settings
    """
    if ui_exposure_settings is None:
        return NullUIRules()

    rules_dict = ui_exposure_settings.pop("rules", None)
    ui_exposure_settings["rules"] = []
    rules: List[Union[UIRules, Rule]] = []

    if len(rules_dict) == 0:
        return NullUIRules()

    for rule in rules_dict:
        rule_type = str(rule.get(metadata_keys.TYPE, ""))
        if rule_type == RuleElementMapping.UI_RULES.name:
            rules.append(construct_ui_rules_from_dict(rule))
        elif rule_type == RuleElementMapping.RULE.name:
            rules.append(construct_attrib_from_dict(rule))
        else:
            raise ValueError(
                f"Invalid UI exposure settings passed to parser. Configuration contains invalid rule "
                f"type: {rule_type}"
            )
    ui_rules: UIRules = construct_attrib_from_dict(ui_exposure_settings)
    ui_rules.rules = rules
    return ui_rules


def create_default_configurable_enum_from_dict(parameter_dict: Union[dict, DictConfig]) -> dict:
    """Create a default configurable enum from a dictionary.

    Takes a parameter_dict representing a configurable Enum and consumes the ENUM_NAME, OPTIONS and DEFAULT_VALUE
    metadata.

    From this, a new subclass of ConfigurableEnum is constructed. The DEFAULT_VALUE of the parameter_dict is updated
    with the appropriate ConfigurableEnum instance.

    Args:
        parameter_dict: Dictionary representation of an enum_selectable parameter. Can be either serialized or not.

    Returns:
        parameter_dict containing the default_value instantiated as a ConfigurableEnum
    """
    # Some input type validation here first, to keep mypy happy
    if isinstance(parameter_dict, DictConfig):
        param_dict = OmegaConf.to_container(parameter_dict)
        if isinstance(param_dict, dict):
            parameter_dict = param_dict
        else:
            raise TypeError(f"Invalid input parameter_dict of type {type(parameter_dict)}")

    # Create the enum using the functional Enum API. Unfortunately this doesn't play nice with mypy, so ignoring the
    # type error for now
    # pylint: disable=unexpected-keyword-arg
    configurable_enum = ConfigurableEnum(
        parameter_dict.pop(metadata_keys.ENUM_NAME),
        names=parameter_dict.pop(metadata_keys.OPTIONS),
    )  # type: ignore
    # pylint: enable=unexpected-keyword-arg
    serialized_default_value = parameter_dict.pop(metadata_keys.DEFAULT_VALUE)

    if isinstance(serialized_default_value, ConfigurableEnum):
        default_value = serialized_default_value
    else:
        # This instantiates the enum, but mypy doesn't pick up that the created configurable_enum type is callable.
        default_value = configurable_enum(serialized_default_value)  # type: ignore

    parameter_dict.update({metadata_keys.DEFAULT_VALUE: default_value})
    return parameter_dict


def gather_parameter_arguments_and_values_from_dict(config_dict_section: Union[dict, DictConfig]) -> dict:
    """Collect arguments needed to construct attrs class out of a config dict section representing a parameter group.

    Parameters living in the group are constructed in this function as well.

    This function returns a dictionary that contains the keys `make_arguments`, `call_arguments` and `values`.
    make_arguments are the arguments that should be passed to attr.make_class to dynamically generate a class
        constructor
    call_arguments are the arguments that should be passed in the initialization call to the constructor
    values are the parameter values, that can be set once the instance of the parameter group is created.

    Args:
        config_dict_section: Dictionary representation of a parameter
            group in a configuration, for which to gather the
            constructor arguments

    Returns:
        dictionary containing the make_arguments, call_arguments and
        values parsed from the config_dict_section
    """
    # pylint: disable=too-many-locals
    make_arguments = {}
    call_arguments: dict = {}
    all_parameter_values = {}
    for key, value in config_dict_section.items():
        if isinstance(value, (DictConfig, dict)):
            # In case of a nested dict, value represents the settings for a parameter. The arguments are parsed from
            # the dict here and passed to the constructor function for the parameter, according to its `type`.
            parameter_dict = copy.deepcopy(value)
            parameter_type = str(parameter_dict.pop(metadata_keys.TYPE, None))
            if parameter_type == str(ConfigElementType.SELECTABLE):
                parameter_dict = create_default_configurable_enum_from_dict(parameter_dict)
            elif parameter_type == str(None):
                raise ValueError(
                    f"No type was specified for the configurable " f"parameter or parameter group named '{key}'"
                )
            parameter_value = parameter_dict.pop("value", None)

            metadata_enums: Dict[str, Enum] = {}
            for metadata_key, enum_type in METADATA_ENUMS.items():
                enum_value = parameter_dict.pop(metadata_key, None)
                if enum_value is not None:
                    metadata_enums.update({metadata_key: deserialize_enum_value(enum_value, enum_type=enum_type)})

            parameter_ui_rules_dict = parameter_dict.pop(metadata_keys.UI_RULES, None)
            parameter_constructor = PrimitiveElementMapping[parameter_type].value
            parameter_ui_rules = construct_ui_rules_from_dict(parameter_ui_rules_dict)
            parameter_make_arguments = {
                key: parameter_constructor(**parameter_dict, ui_rules=parameter_ui_rules, **metadata_enums)
            }
            make_arguments.update(parameter_make_arguments)
            all_parameter_values.update({key: parameter_value})
        else:
            # Flat values will be passed to the initialization call directly. The only exception is the 'type' item,
            # this should not be included in the call_arguments as it is used to determine the type of constructor for
            # attrs to generate
            if key != metadata_keys.TYPE:
                call_arguments.update({key: value})
    return {
        "make_arguments": make_arguments,
        "call_arguments": call_arguments,
        "values": all_parameter_values,
    }


def create_parameter_group(config_dict_section: Union[dict, DictConfig]) -> ParameterGroupTypeVar:
    """Creates a parameter group object out of a config_dict_section.

    config_dict_section is a dictionary or DictConfig representing a parameter group.
    This method should only be used for simple groups, i.e. parameter groups not containing any other parameter groups.
    For nested groups, the function 'create_nested_parameter_group' should be used instead.

    Args:
        config_dict_section: Dictionary representation of the parameter
            group to construct

    Returns:
        ParameterGroup or ConfigurableParameters object constructed
        according to config_dict_section
    """
    params_and_values = gather_parameter_arguments_and_values_from_dict(config_dict_section)
    make_arguments = params_and_values["make_arguments"]
    call_arguments = params_and_values["call_arguments"]
    all_parameter_values = params_and_values["values"]
    group_type = str(config_dict_section.pop(metadata_keys.TYPE))

    group_constructor_type = GroupElementMapping[group_type].value
    group_constructor = attr.make_class(
        GroupElementMapping[group_type].name,
        bases=(group_constructor_type,),
        attrs=make_arguments,
        eq=False,
        order=False,
    )

    parameter_group = group_constructor(**call_arguments)

    for parameter, value in all_parameter_values.items():
        if value is not None:
            setattr(parameter_group, parameter, value)

    parameter_group.update_auto_hpo_states()
    return parameter_group


def create_nested_parameter_group(config_dict_section: Union[dict, DictConfig]) -> ParameterGroup:
    """Creates a parameter group object out of a config_dict_section.

    config_dict_section is a dictionary or DictConfig representing a parameter group. This method should be used for
    nested groups, and uses recursion to reconstruct those.

    Args:
        config_dict_section: Dictionary representation of the parameter group to construct

    Returns:
        ParameterGroup or Configuration object constructed according to config_dict_section
    """
    groups = {}
    main_config = copy.deepcopy(config_dict_section)

    group_names = contains_parameter_groups(config_dict_section)
    for group_name in group_names:
        group_config_section = config_dict_section.pop(group_name)

        if not contains_parameter_groups(group_config_section):
            childless_parameter_group: ParameterGroup = create_parameter_group(group_config_section)
            groups.update({group_name: childless_parameter_group})

        else:
            parameter_group_with_children: ParameterGroup = create_nested_parameter_group(group_config_section)
            groups.update({group_name: parameter_group_with_children})

        main_config.pop(group_name, None)

    parameter_group: ParameterGroup = create_parameter_group(main_config)

    for group_name, group in groups.items():
        setattr(parameter_group, group_name, group)

    parameter_group.__attrs_post_init__()

    return parameter_group


def contains_parameter_groups(config_dict: Union[dict, DictConfig]) -> List[str]:
    """Checks whether a configuration or configuration section specified in `config_dict` contains parameter groups.

    Returns a list of the group names if it does, and an empty list otherwise

    Args:
        config_dict: Dictionary or DictConfig representing a configuration or configuration section.

    Returns:
        List of names of parameter groups that are defined in the config_dict, if any. Empty list otherwise.
    """
    if isinstance(config_dict, DictConfig):
        input_dict: dict = OmegaConf.to_container(config_dict)  # type: ignore
        # The config_dict will always be converted to a dict by the OmegaConf.to_container call, but this is not
        # reflected by the return type signature. Therefore, mypy can ignore this error
    else:
        input_dict = config_dict

    groups: List[str] = []
    group_category_types: List[str] = [str(x) for x in ConfigElementType if x.category == ElementCategory.GROUPS]

    for field_name, field_value in input_dict.items():
        if isinstance(field_value, dict):
            # If the field is a dict, check whether the type is a parameter group (or derived from it)
            _type = field_value.get(metadata_keys.TYPE, None)
            if str(_type) in group_category_types:
                groups.append(field_name)
    return groups


def from_dict_attr(config_dict: Union[dict, DictConfig]) -> ConfigurableParameters:
    """Creates a configuration object from an input config_dict.

    Uses recursion to handle nested parameter groups in the config

    Args:
        config_dict (Union[dict, DictConfig]): Dictionary representation of a TaskConfig,
            ProjectConfig or ComponentConfig

    Returns:
        ConfigurableParameters: ParameterGroup object constructed according to config_dict
    """
    local_config_dict = copy.deepcopy(config_dict)

    # Initialize all groups in the config, and collect them into the config_groups dict to assign them to the final
    # configuration later on
    config_groups: Dict[str, ParameterGroup] = {}

    # Determine all parameter groups living at the first level of the config
    base_config_groups = contains_parameter_groups(config_dict)
    for group_name in base_config_groups:
        group_config_section = local_config_dict.pop(group_name)
        # If the group itself contains groups, it is nested. If not, it's flat. Initialization for nested and flat
        # groups is different, hence we check this.
        if contains_parameter_groups(group_config_section):
            config_groups.update({group_name: create_nested_parameter_group(group_config_section)})
        else:
            config_groups.update({group_name: create_parameter_group(group_config_section)})

    # Collect parameters for the high level config from the config_dict and create the config constructor, using the
    # type defined in the dict
    config: ConfigurableParameters = create_parameter_group(local_config_dict)

    # Add the groups to the config
    for group_name, group in config_groups.items():
        setattr(config, group_name, group)

    # Run post-initialization to set the groups and parameters attributes correctly
    config.__attrs_post_init__()

    return config


def create(input_config: Union[str, DictConfig, dict]) -> ConfigurableParameters:
    """Create a configuration object from a yaml string, yaml file path, dictionary or OmegaConf DictConfig object.

    Args:
        input_config: yaml string, dictionary, DictConfig or filepath
            describing a configuration.

    Returns:
        ConfigurableParameters object
    """
    # Parse input, validate config type and convert to dict if needed
    config_dict = input_to_config_dict(copy.deepcopy(input_config))
    # Create config from the resulting dictionary
    config: ConfigurableParameters = from_dict_attr(config_dict)

    return config



"""Unit Tests for the MPA Dataset Pipeline Torchvision to MMDet."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import numpy as np
import pytest
import torch
from PIL import Image, ImageFilter
from torch import Tensor

from otx.mpa.modules.datasets.pipelines.torchvision2mmdet import (
    BranchImage,
    ColorJitter,
    NDArrayToPILImage,
    NDArrayToTensor,
    PILImageToNDArray,
    RandomApply,
    RandomErasing,
    RandomGaussianBlur,
    RandomGrayscale,
)


@pytest.fixture
def data() -> dict[str, np.ndarray]:
    return {"img": np.ones((256, 256, 3), dtype=np.uint8)}


@pytest.fixture()
def image_tensor() -> Tensor:
    return torch.rand(3, 256, 256)


class TestColorJitter:
    def test_call(self, data: dict[str, np.ndarray]) -> None:
        """Test __call__ method of ColorJitter."""
        transform = ColorJitter()
        outputs = transform(data)
        assert outputs.keys() == data.keys()
        assert np.array_equal(outputs["img"], data["img"])

    def test_repr(self) -> None:
        """Test __repr__ method of ColorJitter."""
        transform = ColorJitter(brightness=0.2)
        assert str(transform) == "ColorJitter(brightness=[0.8, 1.2], contrast=None, saturation=None, hue=None)"


class TestRandomGrayscale:
    def test_random_grayscale(self, image_tensor: Tensor) -> None:
        """Test random grayscale."""
        inputs = {"img": image_tensor}
        pipeline = RandomGrayscale(p=0.5)
        outputs = pipeline.forward(inputs)

        assert isinstance(outputs, dict)
        assert set(outputs.keys()) == set(inputs.keys())
        assert outputs["img"].shape == inputs["img"].shape


class TestRandomErasing:
    def test_random_erasing(self, image_tensor: Tensor) -> None:
        """Test random erasing."""
        transform = RandomErasing(p=1.0, scale=(0.02, 0.33), ratio=(0.3, 3.3), value=0, inplace=False)

        data = {"img": image_tensor}
        transformed_data = transform(data)

        assert "img" in transformed_data
        assert transformed_data["img"].shape == data["img"].shape


class TestRandomGaussianBlur:
    def test_random_gaussian_blur(self) -> None:
        """Test initialization."""
        sigma_min = 0.1
        sigma_max = 2.0
        pipeline = RandomGaussianBlur(sigma_min, sigma_max)
        assert pipeline.sigma_min == sigma_min
        assert pipeline.sigma_max == sigma_max

        # Test forward pass
        inputs = {"img": Image.fromarray(np.zeros((256, 256, 3), dtype=np.uint8))}
        outputs = pipeline(inputs)
        assert outputs.keys() == inputs.keys()
        assert isinstance(outputs["img"], Image.Image)

        # Test that the output image is blurred
        blur_radius = (pipeline.sigma_min + pipeline.sigma_max) / 2
        blurred_image = inputs["img"].filter(ImageFilter.GaussianBlur(radius=blur_radius))
        assert np.array_equal(np.array(outputs["img"]), np.array(blurred_image))

    def test_repr(self) -> None:
        """Test __repr__ method of RandomGaussianBlur."""
        pipeline = RandomGaussianBlur(0.1, 2.0)
        assert repr(pipeline) == "RandomGaussianBlur"


class TestRandomApply:
    def test_random_apply_with(self) -> None:
        """Test RandomApply with a single transform."""
        # Define the transforms to be applied randomly
        transform_cfgs = [
            dict(
                type="ColorJitter",
                brightness=0.4,
                contrast=0.4,
                saturation=0.4,
                hue=0.1,
            )
        ]

        # Create the RandomApply pipeline
        random_apply = RandomApply(transform_cfgs, p=0.0)

        # Define the inputs and expected outputs
        inputs = {"img": Image.fromarray(np.ones((256, 256, 3), dtype=np.uint8))}
        results = random_apply(inputs)
        assert np.allclose(np.array(results["img"]), np.array(inputs["img"]))


class TestNDArrayToTensor:
    def test_ndarray_to_tensor_with_single_channel_image(self, data: dict[str, np.ndarray]) -> None:
        """Test NDArrayToTensor with a single channel image."""
        pipeline = NDArrayToTensor(keys=["img"])
        output = pipeline(data)

        assert output["img"].shape == (3, 256, 256)
        assert isinstance(output["img"], torch.Tensor)


class TestNDArrayToPILImage:
    def test_ndarray_to_pil_conversion(self, data: dict[str, np.ndarray]) -> None:
        """Test NDArrayToPILImage with a three channel image."""
        pipeline = NDArrayToPILImage(keys=["img"])
        output = pipeline(data)

        assert isinstance(output["img"], Image.Image)
        assert output["img"].size == (256, 256)

    def test_rept(self) -> None:
        """Test __repr__ method of NDArrayToPILImage."""
        pipeline = NDArrayToPILImage(keys=["image"])
        assert repr(pipeline) == "NDArrayToPILImage"


class TestPILImageToNDArray:
    def test_call(self, data: dict[str, np.ndarray]) -> None:
        """Test __call__ method of PILImageToNDArray."""
        pipeline = PILImageToNDArray(keys=["image"])
        data = {"image": Image.fromarray(data["img"])}
        output = pipeline(data)

        assert isinstance(output["image"], np.ndarray)
        assert output["image"].shape == (256, 256, 3)

    def test_repr(self) -> None:
        """Test __repr__ method of PILImageToNDArray."""
        pipeline = PILImageToNDArray(keys=["image"])
        assert repr(pipeline) == "PILImageToNDArray"


class TestBranchImage:
    def test_branch_image(self) -> None:
        """Test BranchImage pipeline."""
        # Test data
        data = {"img": "test.jpg", "label": 0, "img_fields": ["img"]}

        # Call the pipeline
        key_map = {"img": "img2", "label": "label2"}
        pipeline = BranchImage(key_map)
        results = pipeline(data)

        # Check that the results have been updated correctly
        assert results["img2"] == "test.jpg"
        assert results["label2"] == 0
        assert "img2" in data["img_fields"]
        assert "label2" not in data["img_fields"]

    def test_repr(self) -> None:
        """Test __repr__ method of BranchImage."""
        pipeline = BranchImage()
        assert repr(pipeline) == "BranchImage"



"""Model configuration of Resnet50-MaskRCNN model for Rotated-Detection Task."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# pylint: disable=invalid-name

_base_ = [
    "../../../../common/adapters/mmcv/configs/backbones/resnet50.yaml",
    "../../base/models/detector.py",
]

task = "instance-segmentation"

model = dict(
    type="CustomMaskRCNN",  # Use CustomMaskRCNN for Incremental Learning
    neck=dict(
        type="FPN",
        in_channels=[256, 512, 1024, 2048],
        out_channels=256,
        num_outs=5,
    ),
    rpn_head=dict(
        type="RPNHead",
        in_channels=256,
        feat_channels=256,
        anchor_generator=dict(
            type="AnchorGenerator",
            scales=[8],
            ratios=[0.5, 1.0, 2.0],
            strides=[4, 8, 16, 32, 64],
        ),
        bbox_coder=dict(
            type="DeltaXYWHBBoxCoder",
            target_means=[0.0, 0.0, 0.0, 0.0],
            target_stds=[1.0, 1.0, 1.0, 1.0],
        ),
        loss_cls=dict(type="CrossEntropyLoss", use_sigmoid=True, loss_weight=1.0),
        loss_bbox=dict(type="L1Loss", loss_weight=1.0),
    ),
    roi_head=dict(
        type="CustomRoIHead",  # Use CustomROIHead for Ignore mode
        bbox_roi_extractor=dict(
            type="SingleRoIExtractor",
            roi_layer=dict(type="RoIAlign", output_size=7, sampling_ratio=0),
            out_channels=256,
            featmap_strides=[4, 8, 16, 32],
        ),
        bbox_head=dict(
            type="Shared2FCBBoxHead",
            in_channels=256,
            fc_out_channels=1024,
            roi_feat_size=7,
            num_classes=80,
            bbox_coder=dict(
                type="DeltaXYWHBBoxCoder",
                target_means=[0.0, 0.0, 0.0, 0.0],
                target_stds=[0.1, 0.1, 0.2, 0.2],
            ),
            reg_class_agnostic=False,
            loss_cls=dict(type="CrossEntropyLoss", use_sigmoid=False, loss_weight=1.0),
            loss_bbox=dict(type="L1Loss", loss_weight=1.0),
        ),
        mask_roi_extractor=dict(
            type="SingleRoIExtractor",
            roi_layer=dict(type="RoIAlign", output_size=14, sampling_ratio=0),
            out_channels=256,
            featmap_strides=[4, 8, 16, 32],
        ),
        mask_head=dict(
            type="FCNMaskHead",
            num_convs=4,
            in_channels=256,
            conv_out_channels=256,
            num_classes=80,
            loss_mask=dict(type="CrossEntropyLoss", use_mask=True, loss_weight=1.0),
        ),
    ),
    train_cfg=dict(
        rpn=dict(
            assigner=dict(
                type="MaxIoUAssigner",
                pos_iou_thr=0.7,
                neg_iou_thr=0.3,
                min_pos_iou=0.3,
                match_low_quality=True,
                ignore_iof_thr=-1,
                gpu_assign_thr=300,
            ),
            sampler=dict(
                type="RandomSampler",
                num=256,
                pos_fraction=0.5,
                neg_pos_ub=-1,
                add_gt_as_proposals=False,
            ),
            allowed_border=-1,
            pos_weight=-1,
            debug=False,
        ),
        rpn_proposal=dict(
            nms_across_levels=False,
            nms_pre=2000,
            max_per_img=1000,
            nms=dict(type="nms", iou_threshold=0.7),
            min_bbox_size=0,
        ),
        rcnn=dict(
            assigner=dict(
                type="MaxIoUAssigner",
                pos_iou_thr=0.5,
                neg_iou_thr=0.5,
                min_pos_iou=0.5,
                match_low_quality=True,
                ignore_iof_thr=-1,
                gpu_assign_thr=300,
            ),
            sampler=dict(
                type="RandomSampler",
                num=512,
                pos_fraction=0.25,
                neg_pos_ub=-1,
                add_gt_as_proposals=True,
            ),
            mask_size=28,
            pos_weight=-1,
            debug=False,
        ),
    ),
    test_cfg=dict(
        rpn=dict(
            nms_across_levels=False,
            nms_pre=1000,
            max_per_img=1000,
            nms=dict(type="nms", iou_threshold=0.7),
            min_bbox_size=0,
        ),
        rcnn=dict(
            score_thr=0.05,
            nms=dict(type="nms", iou_threshold=0.5, max_num=100),
            max_per_img=100,
            mask_thr_binary=0.5,
        ),
    ),
)
load_from = "https://download.openmmlab.com/mmdetection/\
v2.0/mask_rcnn/mask_rcnn_r50_fpn_mstrain-poly_3x_coco/\
mask_rcnn_r50_fpn_mstrain-poly_3x_coco_20210524_201154-21b550bb.pth"

evaluation = dict(interval=1, metric="mAP", save_best="mAP", iou_thr=[0.5])
ignore = True



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from mmcv.utils import ConfigDict

from otx.algorithms.detection.adapters.mmdet.utils.builder import build_detector
from otx.mpa.stage import Stage
from otx.mpa.utils.config_utils import recursively_update_cfg, update_or_add_custom_hook
from otx.mpa.utils.logger import get_logger

logger = get_logger()


class DetectionStage(Stage):
    """Patch config to support otx train."""

    MODEL_BUILDER = build_detector

    def configure(self, model_cfg, model_ckpt, data_cfg, training=True, **kwargs):
        """Create MMCV-consumable config from given inputs"""
        logger.info(f"configure!: training={training}")

        cfg = self.cfg
        self.configure_model(cfg, model_cfg, training, **kwargs)
        self.configure_ckpt(cfg, model_ckpt, kwargs.get("pretrained", None))
        self.configure_data(cfg, training, data_cfg, **kwargs)
        self.configure_regularization(cfg, training)
        self.configure_hyperparams(cfg, training, **kwargs)
        self.configure_task(cfg, training, **kwargs)
        self.configure_hook(cfg)
        return cfg

    def configure_model(self, cfg, model_cfg, training, **kwargs):  # noqa: C901
        """Patch config's model.
        Replace cfg.model to model_cfg
        Change model type to super type
        Patch for OMZ backbones
        """
        if model_cfg:
            if hasattr(model_cfg, "model"):
                cfg.merge_from_dict(model_cfg._cfg_dict)
            else:
                raise ValueError(
                    "Unexpected config was passed through 'model_cfg'. "
                    "it should have 'model' attribute in the config"
                )
            cfg.model_task = cfg.model.pop("task", "detection")
            if cfg.model_task != "detection":
                raise ValueError(f"Given model_cfg ({model_cfg.filename}) is not supported by detection recipe")

        super_type = cfg.model.pop("super_type", None)
        if super_type:
            cfg.model.arch_type = cfg.model.type
            cfg.model.type = super_type

        # OV-plugin
        ir_model_path = kwargs.get("ir_model_path")
        if ir_model_path:

            def is_mmov_model(k, v):
                if k == "type" and v.startswith("MMOV"):
                    return True
                return False

            ir_weight_path = kwargs.get("ir_weight_path", None)
            ir_weight_init = kwargs.get("ir_weight_init", False)
            recursively_update_cfg(
                cfg,
                is_mmov_model,
                {"model_path": ir_model_path, "weight_path": ir_weight_path, "init_weight": ir_weight_init},
            )

    def configure_data(self, cfg, training, data_cfg, **kwargs):  # noqa: C901
        """Patch cfg.data.
        Merge cfg and data_cfg
        Match cfg.data.train.type to super_type
        Patch for unlabeled data path ==> This may be moved to SemiDetectionStage
        """
        if data_cfg:
            cfg.merge_from_dict(data_cfg)

        super().configure_data(cfg, training, **kwargs)
        super_type = cfg.data.train.pop("super_type", None)
        if super_type:
            cfg.data.train.org_type = cfg.data.train.type
            cfg.data.train.type = super_type

    def configure_regularization(self, cfg, training):  # noqa: C901
        """Patch regularization parameters."""
        if training:
            if cfg.model.get("l2sp_weight", 0.0) > 0.0:
                logger.info("regularization config!!!!")

                # Checkpoint
                l2sp_ckpt = cfg.model.get("l2sp_ckpt", None)
                if l2sp_ckpt is None:
                    if "pretrained" in cfg.model:
                        l2sp_ckpt = cfg.model.pretrained
                    if cfg.load_from:
                        l2sp_ckpt = cfg.load_from
                cfg.model.l2sp_ckpt = l2sp_ckpt

                # Disable weight decay
                if "weight_decay" in cfg.optimizer:
                    cfg.optimizer.weight_decay = 0.0

    def configure_hyperparams(self, cfg, training, **kwargs):
        """Patch optimization hyparms such as batch size, learning rate."""
        if "hyperparams" in cfg:
            hyperparams = kwargs.get("hyperparams", None)
            if hyperparams is not None:
                bs = hyperparams.get("bs", None)
                if bs is not None:
                    cfg.data.samples_per_gpu = bs

                lr = hyperparams.get("lr", None)
                if lr is not None:
                    cfg.optimizer.lr = lr

    def configure_task(self, cfg, training, **kwargs):
        """Patch config to support training algorithm."""
        if "task_adapt" in cfg:
            logger.info(f"task config!!!!: training={training}")
            self.task_adapt_type = cfg["task_adapt"].get("type", None)
            self.task_adapt_op = cfg["task_adapt"].get("op", "REPLACE")
            self.configure_classes(cfg)

            if self.data_classes != self.model_classes:
                self.configure_task_data_pipeline(cfg)
            # TODO[JAEGUK]: configure_anchor is not working
            if cfg["task_adapt"].get("use_mpa_anchor", False):
                self.configure_anchor(cfg)
            if self.task_adapt_type == "mpa":
                self.configure_bbox_head(cfg)
                self.configure_ema(cfg)
            else:
                src_data_cfg = self.get_data_cfg(cfg, "train")
                src_data_cfg.pop("old_new_indices", None)

    def configure_classes(self, cfg):
        """Patch classes for model and dataset."""
        org_model_classes = self.get_model_classes(cfg)
        data_classes = self.get_data_classes(cfg)

        # Model classes
        if self.task_adapt_op == "REPLACE":
            if len(data_classes) == 0:
                model_classes = org_model_classes.copy()
            else:
                model_classes = data_classes.copy()
        elif self.task_adapt_op == "MERGE":
            model_classes = org_model_classes + [cls for cls in data_classes if cls not in org_model_classes]
        else:
            raise KeyError(f"{self.task_adapt_op} is not supported for task_adapt options!")

        if self.task_adapt_type == "mpa":
            data_classes = model_classes
        cfg.task_adapt.final = model_classes
        cfg.model.task_adapt = ConfigDict(
            src_classes=org_model_classes,
            dst_classes=model_classes,
        )

        # Model architecture
        head_names = ("mask_head", "bbox_head", "segm_head")
        num_classes = len(model_classes)
        if "roi_head" in cfg.model:
            # For Faster-RCNNs
            for head_name in head_names:
                if head_name in cfg.model.roi_head:
                    if isinstance(cfg.model.roi_head[head_name], list):
                        for head in cfg.model.roi_head[head_name]:
                            head.num_classes = num_classes
                    else:
                        cfg.model.roi_head[head_name].num_classes = num_classes
        else:
            # For other architectures (including SSD)
            for head_name in head_names:
                if head_name in cfg.model:
                    cfg.model[head_name].num_classes = num_classes

        # Eval datasets
        if cfg.get("task", "detection") == "detection":
            eval_types = ["val", "test"]
            for eval_type in eval_types:
                if cfg.data[eval_type]["type"] == "TaskAdaptEvalDataset":
                    cfg.data[eval_type]["model_classes"] = model_classes
                else:
                    # Wrap original dataset config
                    org_type = cfg.data[eval_type]["type"]
                    cfg.data[eval_type]["type"] = "TaskAdaptEvalDataset"
                    cfg.data[eval_type]["org_type"] = org_type
                    cfg.data[eval_type]["model_classes"] = model_classes

        self.org_model_classes = org_model_classes
        self.model_classes = model_classes
        self.data_classes = data_classes

    def configure_task_data_pipeline(self, cfg):
        # Trying to alter class indices of training data according to model class order
        tr_data_cfg = self.get_data_cfg(cfg, "train")
        class_adapt_cfg = dict(type="AdaptClassLabels", src_classes=self.data_classes, dst_classes=self.model_classes)
        pipeline_cfg = tr_data_cfg.pipeline
        for i, op in enumerate(pipeline_cfg):
            if op["type"] == "LoadAnnotations":  # insert just after this op
                op_next_ann = pipeline_cfg[i + 1] if i + 1 < len(pipeline_cfg) else {}
                if op_next_ann.get("type", "") == class_adapt_cfg["type"]:
                    op_next_ann.update(class_adapt_cfg)
                else:
                    pipeline_cfg.insert(i + 1, class_adapt_cfg)
                break

    def configure_anchor(self, cfg, proposal_ratio=None):
        if cfg.model.type in ["SingleStageDetector", "CustomSingleStageDetector"]:
            anchor_cfg = cfg.model.bbox_head.anchor_generator
            if anchor_cfg.type == "SSDAnchorGeneratorClustered":
                cfg.model.bbox_head.anchor_generator.pop("input_size", None)

    def configure_bbox_head(self, cfg):
        """Patch bbox head in detector for class incremental learning.
        Most of patching are related with hyper-params in focal loss
        """
        if cfg.get("task", "detection") == "detection":
            bbox_head = cfg.model.bbox_head
        else:
            bbox_head = cfg.model.roi_head.bbox_head

        alpha, gamma = 0.25, 2.0
        if bbox_head.type in ["SSDHead", "CustomSSDHead"]:
            gamma = 1 if cfg["task_adapt"].get("efficient_mode", False) else 2
            bbox_head.type = "CustomSSDHead"
            bbox_head.loss_cls = ConfigDict(
                type="FocalLoss",
                loss_weight=1.0,
                gamma=gamma,
                reduction="none",
            )
        elif bbox_head.type in ["ATSSHead"]:
            gamma = 3 if cfg["task_adapt"].get("efficient_mode", False) else 4.5
            bbox_head.loss_cls.gamma = gamma
        elif bbox_head.type in ["VFNetHead", "CustomVFNetHead"]:
            alpha = 0.75
            gamma = 1 if cfg["task_adapt"].get("efficient_mode", False) else 2
        # TODO Move this part
        # This is not related with patching bbox head
        elif bbox_head.type in ["YOLOXHead", "CustomYOLOXHead"]:
            if cfg.data.train.type == "MultiImageMixDataset":
                self.add_yolox_hooks(cfg)

        if cfg.get("ignore", False):
            bbox_head.loss_cls = ConfigDict(
                type="CrossSigmoidFocalLoss",
                use_sigmoid=True,
                num_classes=len(self.model_classes),
                alpha=alpha,
                gamma=gamma,
            )

    @staticmethod
    def configure_ema(cfg):
        """Patch ema settings."""
        adaptive_ema = cfg.get("adaptive_ema", {})
        if adaptive_ema:
            update_or_add_custom_hook(
                cfg,
                ConfigDict(
                    type="CustomModelEMAHook",
                    priority="ABOVE_NORMAL",
                    resume_from=cfg.get("resume_from"),
                    **adaptive_ema,
                ),
            )
        else:
            update_or_add_custom_hook(
                cfg,
                ConfigDict(type="EMAHook", priority="ABOVE_NORMAL", resume_from=cfg.get("resume_from"), momentum=0.1),
            )

    @staticmethod
    def add_yolox_hooks(cfg):
        update_or_add_custom_hook(
            cfg,
            ConfigDict(
                type="YOLOXModeSwitchHook",
                num_last_epochs=15,
                priority=48,
            ),
        )
        update_or_add_custom_hook(
            cfg,
            ConfigDict(
                type="SyncNormHook",
                num_last_epochs=15,
                interval=1,
                priority=48,
            ),
        )



import pytest

from otx.api.configuration.configurable_parameters import ConfigurableParameters
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.model import ModelConfiguration, ModelEntity


@pytest.fixture
def otx_model():
    model_configuration = ModelConfiguration(
        configurable_parameters=ConfigurableParameters(header="header", description="description"),
        label_schema=LabelSchemaEntity(),
    )
    return ModelEntity(train_dataset=DatasetEntity(), configuration=model_configuration)



"""Initialization of demo package."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .executors import AsyncExecutor, ChainExecutor, SyncExecutor
from .model_container import ModelContainer
from .utils import create_output_converter, create_visualizer

__all__ = [
    "SyncExecutor",
    "AsyncExecutor",
    "ChainExecutor",
    "create_output_converter",
    "create_visualizer",
    "ModelContainer",
]



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import random
import re
from copy import deepcopy

import numpy as np
from mmcls.datasets.builder import PIPELINES
from PIL import Image

from otx.mpa.modules.datasets.pipelines.transforms.augments import CythonAugments

_AUGMIX_TRANSFORMS_GREY = [
    "SharpnessIncreasing",  # not in paper
    "ShearX",
    "ShearY",
    "TranslateXRel",
    "TranslateYRel",
]

_AUGMIX_TRANSFORMS = [
    "AutoContrast",
    "ColorIncreasing",  # not in paper
    "ContrastIncreasing",  # not in paper
    "BrightnessIncreasing",  # not in paper
    "SharpnessIncreasing",  # not in paper
    "Equalize",
    "PosterizeIncreasing",
    "SolarizeIncreasing",
    "ShearX",
    "ShearY",
    "TranslateXRel",
    "TranslateYRel",
]


class OpsFabric:
    def __init__(self, name, magnitude, hparams, prob=1.0):
        self.max_level = 10
        self.prob = prob
        self.hparams = hparams
        # kwargs for augment functions
        self.aug_kwargs = dict(fillcolor=hparams["img_mean"], resample=(Image.BILINEAR, Image.BICUBIC))
        self.LEVEL_TO_ARG = {
            "AutoContrast": None,
            "Equalize": None,
            "Rotate": self._rotate_level_to_arg,
            "PosterizeIncreasing": self._posterize_increasing_level_to_arg,
            "SolarizeIncreasing": self._solarize_increasing_level_to_arg,
            "ColorIncreasing": self._enhance_increasing_level_to_arg,
            "ContrastIncreasing": self._enhance_increasing_level_to_arg,
            "BrightnessIncreasing": self._enhance_increasing_level_to_arg,
            "SharpnessIncreasing": self._enhance_increasing_level_to_arg,
            "ShearX": self._shear_level_to_arg,
            "ShearY": self._shear_level_to_arg,
            "TranslateXRel": self._translate_rel_level_to_arg,
            "TranslateYRel": self._translate_rel_level_to_arg,
        }
        self.NAME_TO_OP = {
            "AutoContrast": CythonAugments.autocontrast,
            "Equalize": CythonAugments.equalize,
            "Rotate": CythonAugments.rotate,
            "PosterizeIncreasing": CythonAugments.posterize,
            "SolarizeIncreasing": CythonAugments.solarize,
            "ColorIncreasing": CythonAugments.color,
            "ContrastIncreasing": CythonAugments.contrast,
            "BrightnessIncreasing": CythonAugments.brightness,
            "SharpnessIncreasing": CythonAugments.sharpness,
            "ShearX": CythonAugments.shear_x,
            "ShearY": CythonAugments.shear_y,
            "TranslateXRel": CythonAugments.translate_x_rel,
            "TranslateYRel": CythonAugments.translate_y_rel,
        }
        self.aug_fn = self.NAME_TO_OP[name]
        self.level_fn = self.LEVEL_TO_ARG[name]
        self.magnitude = magnitude
        self.magnitude_std = self.hparams.get("magnitude_std", float("inf"))

    @staticmethod
    def randomly_negate(v):
        """With 50% prob, negate the value"""
        return -v if random.random() > 0.5 else v

    def _rotate_level_to_arg(self, level, _hparams):
        # range [-30, 30]
        level = (level / self.max_level) * 30.0
        level = self.randomly_negate(level)
        return (level,)

    def _enhance_increasing_level_to_arg(self, level, _hparams):
        # range [0.1, 1.9]
        level = (level / self.max_level) * 0.9
        level = 1.0 + self.randomly_negate(level)
        return (level,)

    def _shear_level_to_arg(self, level, _hparams):
        # range [-0.3, 0.3]
        level = (level / self.max_level) * 0.3
        level = self.randomly_negate(level)
        return (level,)

    def _translate_rel_level_to_arg(self, level, hparams):
        # default range [-0.45, 0.45]
        translate_pct = hparams.get("translate_pct", 0.45)
        level = (level / self.max_level) * translate_pct
        level = self.randomly_negate(level)
        return (level,)

    def _posterize_level_to_arg(self, level, _hparams):
        # range [0, 4], 'keep 0 up to 4 MSB of original image'
        # intensity/severity of augmentation decreases with level
        return (int((level / self.max_level) * 4),)

    def _posterize_increasing_level_to_arg(self, level, hparams):
        # range [4, 0], 'keep 4 down to 0 MSB of original image',
        # intensity/severity of augmentation increases with level
        return (4 - self._posterize_level_to_arg(level, hparams)[0],)

    def _solarize_level_to_arg(self, level, _hparams):
        # range [0, 256]
        # intensity/severity of augmentation decreases with level
        return (int((level / self.max_level) * 256),)

    def _solarize_increasing_level_to_arg(self, level, _hparams):
        # range [0, 256]
        # intensity/severity of augmentation increases with level
        return (256 - self._solarize_level_to_arg(level, _hparams)[0],)

    def __call__(self, img):
        if self.prob < 1.0 and random.random() > self.prob:
            return img
        magnitude = self.magnitude
        if self.magnitude_std:
            if self.magnitude_std == float("inf"):
                magnitude = random.uniform(0, magnitude)
            elif self.magnitude_std > 0:
                magnitude = random.gauss(magnitude, self.magnitude_std)
        magnitude = min(self.max_level, max(0, magnitude))  # clip to valid range
        level_args = self.level_fn(magnitude, self.hparams) if self.level_fn is not None else tuple()
        return self.aug_fn(img, *level_args, **self.aug_kwargs)


@PIPELINES.register_module()
class AugMixAugment(object):
    """AugMix Transform
    Adapted and improved from impl here: https://github.com/google-research/augmix/blob/master/imagenet.py
    From paper: 'AugMix: A Simple Data Processing Method to Improve Robustness and Uncertainty -
    https://arxiv.org/abs/1912.02781
    """

    def __init__(self, config_str, image_mean=None, grey=False, **kwargs):
        self.ops, self.alpha, self.width, self.depth = self._augmix_ops(config_str, image_mean, grey=grey)

    def _apply_basic(self, img, mixing_weights, m):
        # This is a literal adaptation of the paper/official implementation without normalizations and
        # PIL <-> Numpy conversions between every op. It is still quite CPU compute heavy compared to the
        # typical augmentation transforms, could use a GPU / Kornia implementation.
        mixed = (1 - m) * np.array(img, dtype=np.float32)
        for mw in mixing_weights:
            depth = self.depth if self.depth > 0 else np.random.randint(1, 4)
            ops = np.random.choice(self.ops, depth, replace=True)
            img_aug = deepcopy(img)
            for op in ops:
                img_aug = op(img_aug)
            CythonAugments.blend(img_aug, mixed, mw * m)
        np.clip(mixed, 0, 255.0, out=mixed)
        return Image.fromarray(mixed.astype(np.uint8))

    def _augmix_ops(self, config_str, image_mean=None, translate_const=250, grey=False):
        if image_mean is None:
            image_mean = [0.485, 0.456, 0.406]  # imagenet mean
        magnitude = 3
        width = 3
        depth = -1
        alpha = 1.0
        p = 1.0
        hparams = dict(
            translate_const=translate_const,
            img_mean=tuple([int(c * 256) for c in image_mean]),
            magnitude_std=float("inf"),
        )
        config = config_str.split("-")
        assert config[0] == "augmix"
        config = config[1:]
        for c in config:
            cs = re.split(r"(\d.*)", c)
            if len(cs) < 2:
                continue
            key, val = cs[:2]
            if key == "mstd":
                hparams.setdefault("magnitude_std", float(val))
            elif key == "m":
                magnitude = int(val)
            elif key == "w":
                width = int(val)
            elif key == "d":
                depth = int(val)
            elif key == "a":
                alpha = float(val)
            elif key == "p":
                p = float(val)
            else:
                assert False, "Unknown AugMix config section"
        aug_politics = _AUGMIX_TRANSFORMS_GREY if grey else _AUGMIX_TRANSFORMS
        return (
            [OpsFabric(name, magnitude, hparams, p) for name in aug_politics],
            alpha,
            width,
            depth,
        )

    def __call__(self, results):
        for key in results.get("img_fields", ["img"]):
            img = results[key]
            if not Image.isImageType(img):
                img = Image.fromarray(img)
            mixing_weights = np.float32(np.random.dirichlet([self.alpha] * self.width))
            m = np.float32(np.random.beta(self.alpha, self.alpha))
            mixed = self._apply_basic(img, mixing_weights, m)
            results["augmix"] = True
            results[key] = mixed
        return results



"""Wrapper for Open Model Zoo for Anomaly Classification tasks."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Any, Dict

import cv2
import numpy as np

from .base import AnomalyBase


class AnomalyClassification(AnomalyBase):
    """Wrapper for anomaly classification task."""

    __model__ = "anomaly_classification"

    def postprocess(self, outputs: Dict[str, np.ndarray], meta: Dict[str, Any]) -> float:
        """Resize the outputs of the model to original image size.

        Args:
            outputs (Dict[str, np.ndarray]): Raw outputs of the model after ``infer_sync`` is called.
            meta (Dict[str, Any]): Metadata which contains values such as threshold, original image size.

        Returns:
            float: Normalized anomaly score
        """
        anomaly_map: np.ndarray = outputs[self.output_blob_name].squeeze()
        pred_score = anomaly_map.reshape(-1).max()

        meta["image_threshold"] = self.image_threshold  # pylint: disable=no-member
        meta["min"] = self.min  # pylint: disable=no-member
        meta["max"] = self.max  # pylint: disable=no-member
        meta["threshold"] = self.threshold  # pylint: disable=no-member

        anomaly_map = self._normalize(anomaly_map, meta["image_threshold"], meta["min"], meta["max"])
        pred_score = self._normalize(pred_score, meta["image_threshold"], meta["min"], meta["max"])

        input_image_height = meta["original_shape"][0]
        input_image_width = meta["original_shape"][1]
        result = cv2.resize(anomaly_map, (input_image_width, input_image_height))

        meta["anomaly_map"] = result

        return np.array(pred_score)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from otx.mpa.modules.models.losses.l2sp_loss import L2SPLoss


class L2SPDetectorMixin(object):
    """L2SP-enabled detector mix-in"""

    def __init__(self, l2sp_ckpt=None, l2sp_weight=None, **kwargs):
        super(L2SPDetectorMixin, self).__init__(**kwargs)
        if l2sp_ckpt and l2sp_weight:
            self.l2sp = L2SPLoss(self, l2sp_ckpt, l2sp_weight)
            print("L2SP initilaized!")
        else:
            self.l2sp = None

    def forward_train(self, *args, **kwargs):
        losses = super().forward_train(*args, **kwargs)

        # Add L2SP regularization loss
        # (Assuming weight decay is disable in optimizer setting)
        if self.l2sp:
            losses.update(dict(loss_l2sp=self.l2sp()))
        return losses



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from mmcv.utils import Config, ConfigDict
from mmdet.datasets import build_dataloader as mmdet_build_dataloader
from mmdet.datasets import build_dataset as mmdet_build_dataset
from mmdet.datasets import replace_ImageToTensor

from otx.algorithms.common.adapters.mmcv.utils import (
    build_data_parallel,
    build_dataloader,
    build_dataset,
)
from otx.algorithms.detection.adapters.mmdet.data import ImageTilingDataset
from otx.mpa.modules.hooks.recording_forward_hooks import (
    ActivationMapHook,
    DetSaliencyMapHook,
    EigenCamHook,
)
from otx.mpa.registry import STAGES
from otx.mpa.utils.logger import get_logger

from .stage import DetectionStage

logger = get_logger()
EXPLAINER_HOOK_SELECTOR = {
    "classwisesaliencymap": DetSaliencyMapHook,
    "eigencam": EigenCamHook,
    "activationmap": ActivationMapHook,
}


@STAGES.register_module()
class DetectionExplainer(DetectionStage):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dataset = None

    def run(self, model_cfg, model_ckpt, data_cfg, **kwargs):
        """Run explain stage for detection

        - Configuration
        - Environment setup
        - Run inference via MMDetection -> MMCV
        """
        self._init_logger()
        explainer = kwargs.get("explainer")
        self.explainer_hook = EXPLAINER_HOOK_SELECTOR.get(explainer.lower(), None)
        if self.explainer_hook is None:
            raise NotImplementedError(f"Explainer algorithm {explainer} not supported!")
        logger.info(f"Explainer algorithm: {explainer}")

        cfg = self.configure(model_cfg, model_ckpt, data_cfg, training=False, **kwargs)
        logger.info("explain!")

        model_builder = kwargs.get("model_builder", None)
        outputs = self.explain(cfg, model_builder)

        return dict(outputs=outputs)

    def explain(self, cfg, model_builder=None):
        # TODO: distributed inference

        data_cfg = cfg.data.test.copy()

        # Input source
        if "input_source" in cfg:
            input_source = cfg.get("input_source")
            logger.info(f"Inferring on input source: data.{input_source}")
            src_data_cfg = cfg.data[input_source]
            data_cfg.test_mode = src_data_cfg.get("test_mode", False)
            data_cfg.ann_file = src_data_cfg.ann_file
            data_cfg.img_prefix = src_data_cfg.img_prefix
            if "classes" in src_data_cfg:
                data_cfg.classes = src_data_cfg.classes

        data_cfg = Config(
            ConfigDict(
                data=ConfigDict(
                    samples_per_gpu=cfg.data.get("samples_per_gpu", 1),
                    workers_per_gpu=cfg.data.get("workers_per_gpu", 0),
                    test=data_cfg,
                    test_dataloader=cfg.data.get("test_dataloader", {}).copy(),
                ),
                gpu_ids=cfg.gpu_ids,
                seed=cfg.get("seed", None),
                model_task=cfg.model_task,
            )
        )
        self.configure_samples_per_gpu(data_cfg, "test", distributed=False)
        samples_per_gpu = data_cfg.data.test_dataloader.get(
            "samples_per_gpu",
            data_cfg.data.get("samples_per_gpu", 1),
        )
        if samples_per_gpu > 1:
            # Replace 'ImageToTensor' to 'DefaultFormatBundle'
            data_cfg.data.test.pipeline = replace_ImageToTensor(data_cfg.data.test.pipeline)

        # Data loader
        self.dataset = build_dataset(data_cfg, "test", mmdet_build_dataset)
        test_dataloader = build_dataloader(
            self.dataset,
            data_cfg,
            "test",
            mmdet_build_dataloader,
            distributed=False,
        )

        # Target classes
        target_classes = self.dataset.CLASSES
        head_names = ("mask_head", "bbox_head", "segm_head")
        num_classes = len(target_classes)
        if "roi_head" in cfg.model:
            # For Faster-RCNNs
            for head_name in head_names:
                if head_name in cfg.model.roi_head:
                    if isinstance(cfg.model.roi_head[head_name], list):
                        for head in cfg.model.roi_head[head_name]:
                            head.num_classes = num_classes
                    else:
                        cfg.model.roi_head[head_name].num_classes = num_classes
        else:
            # For other architectures (including SSD)
            for head_name in head_names:
                if head_name in cfg.model:
                    cfg.model[head_name].num_classes = num_classes

        # Model
        cfg.model.pretrained = None
        if cfg.model.get("neck"):
            if isinstance(cfg.model.neck, list):
                for neck_cfg in cfg.model.neck:
                    if neck_cfg.get("rfp_backbone"):
                        if neck_cfg.rfp_backbone.get("pretrained"):
                            neck_cfg.rfp_backbone.pretrained = None
            elif cfg.model.neck.get("rfp_backbone"):
                if cfg.model.neck.rfp_backbone.get("pretrained"):
                    cfg.model.neck.rfp_backbone.pretrained = None
        # TODO: Check Inference FP16 Support
        model = self.build_model(cfg, model_builder, fp16=False)
        model.CLASSES = target_classes
        model.eval()
        feature_model = self._get_feature_module(model)
        model = build_data_parallel(model, cfg, distributed=False)

        # InferenceProgressCallback (Time Monitor enable into Infer task)
        self.set_inference_progress_callback(model, cfg)

        # Class-wise Saliency map for Single-Stage Detector, otherwise use class-ignore saliency map.
        eval_predictions = []
        with self.explainer_hook(feature_model) as saliency_hook:
            for data in test_dataloader:
                result = model(return_loss=False, rescale=True, **data)
                eval_predictions.extend(result)
            saliency_maps = saliency_hook.records

        # Check and unwrap ImageTilingDataset object from TaskAdaptEvalDataset
        dataset = self.dataset
        while hasattr(dataset, "dataset") and not isinstance(dataset, ImageTilingDataset):
            dataset = dataset.dataset

        # In the tiling case, select the first images which is map of the entire image
        if isinstance(dataset, ImageTilingDataset):
            saliency_maps = [saliency_maps[i] for i in range(dataset.num_samples)]

        outputs = dict(detections=eval_predictions, saliency_maps=saliency_maps)
        return outputs



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import pytest
import torch

from otx.mpa.modules.models.losses.cross_entropy_loss import CrossEntropyLossWithIgnore
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestCrossEntropyLossWithIgnore:
    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.num_classes = 2
        self.default_gt = torch.zeros(2).long()
        self.default_input = torch.zeros((2, self.num_classes))
        self.default_input[0, 0] = 1
        self.default_loss = CrossEntropyLossWithIgnore()

    @e2e_pytest_unit
    def test_forward(self):
        result_c = self.default_loss(self.default_input, self.default_gt)
        self.default_input[0, 1] = 1
        result_w = self.default_loss(self.default_input, self.default_gt)
        assert result_c < result_w

    def test_weight(self):
        result = self.default_loss(self.default_input, self.default_gt)
        loss_w = CrossEntropyLossWithIgnore(loss_weight=2)
        result_w = loss_w(self.default_input, self.default_gt)
        assert result_w > result

    def test_ignore(self):
        loss_i = CrossEntropyLossWithIgnore(ignore_index=0)
        result_i = loss_i(self.default_input, self.default_gt)
        assert result_i == 0

    def test_reduction(self):
        result = self.default_loss(self.default_input, self.default_gt)
        loss_s = CrossEntropyLossWithIgnore(reduction="sum")
        result_s = loss_s(self.default_input, self.default_gt)
        assert result_s > result



# Copyright (C) 2020-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

"""Create MVTec AD (CC BY-NC-SA 4.0) JSON Annotations for OTX CLI.

Description:
    This script converts MVTec AD dataset masks to OTX CLI annotation format for
        classification, detection and segmentation tasks.

License:
    MVTec AD dataset is released under the Creative Commons
    Attribution-NonCommercial-ShareAlike 4.0 International License
    (CC BY-NC-SA 4.0)(https://creativecommons.org/licenses/by-nc-sa/4.0/).

Reference:
    - Paul Bergmann, Kilian Batzner, Michael Fauser, David Sattlegger, Carsten Steger:
      The MVTec Anomaly Detection Dataset: A Comprehensive Real-World Dataset for
      Unsupervised Anomaly Detection; in: International Journal of Computer Vision
      129(4):1038-1059, 2021, DOI: 10.1007/s11263-020-01400-4.

    - Paul Bergmann, Michael Fauser, David Sattlegger, Carsten Steger: MVTec AD —
      A Comprehensive Real-World Dataset for Unsupervised Anomaly Detection;
      in: IEEE/CVF Conference on Computer Vision and Pattern Recognition (CVPR),
      9584-9592, 2019, DOI: 10.1109/CVPR.2019.00982.

Example:
    Assume that MVTec AD dataset is located in "./data/anomaly/MVTec/" from the root
    directory in training_extensions. JSON annotations could be created by running the
    following:

    >>> import os
    '~/training_extensions'
    >>> os.listdir("./data/anomaly")
    ['detection', 'shapes', 'segmentation', 'MVTec', 'classification']

    The following script will generate the classification, detection and segmentation
    JSON annotations to each category in ./data/anomaly/MVTec dataset.

    >>> python external/anomaly/adapters.anomalib/data/create_mvtec_ad_json_annotations.py \
    ...     --data_path ./data/anomaly/MVTec/
"""

import json
import os
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Any, Dict, List, Optional

import cv2
import pandas as pd
from anomalib.data.mvtec import make_mvtec_dataset
from anomalib.data.utils import Split


def create_bboxes_from_mask(mask_path: str) -> List[List[float]]:
    """Create bounding box from binary mask.

    Args:
        mask_path (str): Path to binary mask.

    Returns:
        List[List[float]]: Bounding box coordinates.
    """
    # pylint: disable-msg=too-many-locals

    mask = cv2.imread(mask_path, cv2.IMREAD_GRAYSCALE)
    height, width = mask.shape

    bboxes: List[List[float]] = []
    _, _, coordinates, _ = cv2.connectedComponentsWithStats(mask)
    for i, coordinate in enumerate(coordinates):
        # First row of the coordinates is always backround,
        # so should be ignored.
        if i == 0:
            continue

        # Last column of the coordinates is the area of the connected component.
        # It could therefore be ignored.
        comp_x, comp_y, comp_w, comp_h, _ = coordinate
        x1 = comp_x / width
        y1 = comp_y / height
        x2 = (comp_x + comp_w) / width
        y2 = (comp_y + comp_h) / height

        bboxes.append([x1, y1, x2, y2])

    return bboxes


def create_polygons_from_mask(mask_path: str) -> List[List[List[float]]]:
    """Create polygons from binary mask.

    Args:
        mask_path (str): Path to binary mask.

    Returns:
        List[List[float]]: Polygon coordinates.
    """
    mask = cv2.imread(mask_path, cv2.IMREAD_GRAYSCALE)
    height, width = mask.shape

    polygons = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)[0]
    polygons = [[[point[0][0] / width, point[0][1] / height] for point in polygon] for polygon in polygons]

    return polygons


def create_classification_json_items(pd_items: pd.DataFrame) -> Dict[str, Any]:
    """Create JSON items for the classification task.

    Args:
        pd_items (pd.DataFrame): MVTec AD samples in pandas DataFrame object.

    Returns:
        Dict[str, Any]: MVTec AD classification JSON items
    """
    json_items: Dict[str, Any] = {"image_path": {}, "label": {}, "masks": {}}
    for index, pd_item in pd_items.iterrows():
        json_items["image_path"][str(index)] = pd_item.image_path.replace(pd_item.path, "")[1:]
        json_items["label"][str(index)] = pd_item.label
        if pd_item.label != "good":
            json_items["masks"][str(index)] = pd_item.mask_path.replace(pd_item.path, "")[1:]

    return json_items


def create_detection_json_items(pd_items: pd.DataFrame) -> Dict[str, Any]:
    """Create JSON items for the detection task.

    Args:
        pd_items (pd.DataFrame): MVTec AD samples in pandas DataFrame object.

    Returns:
        Dict[str, Any]: MVTec AD detection JSON items
    """
    json_items: Dict[str, Any] = {"image_path": {}, "label": {}, "bboxes": {}}
    for index, pd_item in pd_items.iterrows():
        json_items["image_path"][str(index)] = pd_item.image_path.replace(pd_item.path, "")[1:]
        json_items["label"][str(index)] = pd_item.label
        if pd_item.label != "good":
            json_items["bboxes"][str(index)] = create_bboxes_from_mask(pd_item.mask_path)

    return json_items


def create_segmentation_json_items(pd_items: pd.DataFrame) -> Dict[str, Any]:
    """Create JSON items for the segmentation task.

    Args:
        pd_items (pd.DataFrame): MVTec AD samples in pandas DataFrame object.

    Returns:
        Dict[str, Any]: MVTec AD segmentation JSON items
    """
    json_items: Dict[str, Any] = {"image_path": {}, "label": {}, "masks": {}}
    for index, pd_item in pd_items.iterrows():
        json_items["image_path"][str(index)] = pd_item.image_path.replace(pd_item.path, "")[1:]
        json_items["label"][str(index)] = pd_item.label
        if pd_item.label != "good":
            json_items["masks"][str(index)] = create_polygons_from_mask(pd_item.mask_path)

    return json_items


def save_json_items(json_items: Dict[str, Any], file: str) -> None:
    """Save JSON items to file.

    Args:
        json_items (Dict[str, Any]): MVTec AD JSON items
        file (str): Path to save as a JSON file.
    """
    with open(file=file, mode="w", encoding="utf-8") as f:
        json.dump(json_items, f)


def create_task_annotations(task: str, data_path: str, annotation_path: str) -> None:
    """Create MVTec AD categories for a given task.

    Args:
        task (str): Task type to save annotations.
        data_path (str): Path to MVTec AD category.
        annotation_path (str): Path to save MVTec AD category JSON annotation items.

    Raises:
        ValueError: When task is not classification, detection or segmentation.
    """
    annotation_path = os.path.join(annotation_path, task)
    os.makedirs(annotation_path, exist_ok=True)

    for split in ["train", "val", "test"]:

        if task == "classification":
            create_json_items = create_classification_json_items
        elif task == "detection":
            create_json_items = create_detection_json_items
        elif task == "segmentation":
            create_json_items = create_segmentation_json_items
        else:
            raise ValueError(f"Unknown task {task}. Available tasks are classification, detection and segmentation.")

        if split == "train":
            df_items = make_mvtec_dataset(root=Path(data_path), split=Split.TRAIN)
        else:
            df_items = make_mvtec_dataset(root=Path(data_path), split=Split.TEST)
        json_items = create_json_items(df_items)
        save_json_items(json_items, f"{annotation_path}/{split}.json")


def create_mvtec_ad_category_annotations(data_path: str, annotation_path: str) -> None:
    """Create MVTec AD category annotations for classification, detection and segmentation tasks.

    Args:
        data_path (str): Path to MVTec AD category.
        annotation_path (str): Path to save MVTec AD category JSON annotation items.
    """
    for task in ["classification", "detection", "segmentation"]:
        create_task_annotations(task, data_path, annotation_path)


def create_mvtec_ad_annotations(mvtec_data_path: str, mvtec_annotation_path: Optional[str] = None) -> None:
    """Create JSON annotations for MVTec AD dataset.

    Args:
        mvtec_data_path (str): Path to MVTec AD dataset.
        mvtec_annotation_path (Optional[str], optional): Path to save JSON annotations. Defaults to None.
    """
    if mvtec_annotation_path is None:
        mvtec_annotation_path = mvtec_data_path

    categories = [
        "bottle",
        "cable",
        "capsule",
        "carpet",
        "grid",
        "hazelnut",
        "leather",
        "metal_nut",
        "pill",
        "screw",
        "tile",
        "toothbrush",
        "transistor",
        "wood",
        "zipper",
    ]

    for category in categories:
        print(f"Creating annotations for {category}")
        category_data_path = os.path.join(mvtec_data_path, category)
        category_annotation_path = os.path.join(mvtec_annotation_path, category)
        create_mvtec_ad_category_annotations(category_data_path, category_annotation_path)


def get_args() -> Namespace:
    """Get command line arguments.

    Returns:
        Namespace: List of arguments.
    """
    parser = ArgumentParser()
    parser.add_argument("--data_path", type=str, default="./data/anomaly/MVTec/", help="Path to Mvtec AD dataset.")
    parser.add_argument("--annotation_path", type=str, required=False, help="Path to create OTX CLI annotations.")
    return parser.parse_args()


def main():
    """Create MVTec AD Annotations."""
    args = get_args()
    create_mvtec_ad_annotations(mvtec_data_path=args.data_path, mvtec_annotation_path=args.annotation_path)


if __name__ == "__main__":
    main()



"""This module contains the interface class for tasks that can export their models."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import abc
from enum import Enum, auto

from otx.api.entities.model import ModelEntity


class ExportType(Enum):
    """Represent the type of export format available through this interface."""

    OPENVINO = auto()


class IExportTask(metaclass=abc.ABCMeta):
    """A base interface class for tasks which can export their models."""

    @abc.abstractmethod
    def export(self, export_type: ExportType, output_model: ModelEntity, dump_features: bool = False):
        """This method defines the interface for export.

        Args:
            export_type (ExportType): The type of optimization.
            output_model (ModelEntity): The output model entity.
            dump_features (bool): Flag to return "feature_vector" and "saliency_map".
        """
        raise NotImplementedError



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np

from ..builder import SCALAR_SCHEDULERS
from .base import BaseScalarScheduler


@SCALAR_SCHEDULERS.register_module()
class PolyScalarScheduler(BaseScalarScheduler):
    """The learning rate changes over time according to a polynomial schedule.

    Args:
        start_scale (float): The initial learning rate scale.
        end_scale (float): The final learning rate scale.
        num_iters (int): The number of iterations to reach the final learning rate.
        power (float): The power of the polynomial schedule.
        by_epoch (bool): Whether to use epoch as the unit of iteration.
    """

    def __init__(
        self, start_scale: float, end_scale: float, num_iters: int, power: float = 1.2, by_epoch: bool = False
    ):
        super(PolyScalarScheduler, self).__init__()

        self._start_s = start_scale
        assert self._start_s >= 0.0
        self._end_s = end_scale
        assert self._end_s >= 0.0
        self._num_iters = num_iters
        assert self._num_iters >= 0
        self._power = power
        assert self._power >= 0.0
        self.by_epoch = by_epoch

    def _get_value(self, step, epoch_size):
        if step is None:
            return float(self._end_s)

        if self.by_epoch:
            num_iters = epoch_size * self._num_iters
        else:
            num_iters = self._num_iters
        if num_iters == 0:
            return self._end_s

        if step < num_iters:
            factor = (self._end_s - self._start_s) / (1.0 - self._power)
            var_a = factor / (num_iters**self._power)
            var_b = -factor * self._power / float(num_iters)

            out_value = var_a * np.power(step, self._power) + var_b * step + self._start_s
        else:
            out_value = self._end_s

        return out_value



"""BaseTask for NNCF."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.


import io
import json
import os
from collections.abc import Mapping
from copy import deepcopy
from typing import Dict, List, Optional

import torch
from mmcv.utils import ConfigDict

import otx.algorithms.common.adapters.mmcv.nncf.patches  # noqa: F401  # pylint: disable=unused-import
from otx.algorithms.common.adapters.mmcv.utils import (
    get_configs_by_keys,
    remove_from_config,
    remove_from_configs_by_type,
)
from otx.algorithms.common.adapters.nncf import (
    check_nncf_is_enabled,
    is_accuracy_aware_training_set,
)
from otx.algorithms.common.adapters.nncf.config import compose_nncf_config
from otx.algorithms.common.utils.callback import OptimizationProgressCallback
from otx.algorithms.common.utils.data import get_dataset
from otx.api.configuration import cfg_helper
from otx.api.configuration.helper.utils import ids_to_strings
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.model import (
    ModelEntity,
    ModelFormat,
    ModelOptimizationType,
    ModelPrecision,
    OptimizationMethod,
)
from otx.api.entities.optimization_parameters import (
    OptimizationParameters,
    default_progress_callback,
)
from otx.api.entities.subset import Subset
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.serialization.label_mapper import label_schema_to_bytes
from otx.api.usecases.tasks.interfaces.optimization_interface import (
    IOptimizationTask,
    OptimizationType,
)
from otx.api.utils.argument_checks import (
    DatasetParamTypeCheck,
    check_input_parameters_type,
)
from otx.mpa.utils.logger import get_logger

from .training_base import BaseTask

logger = get_logger()


class NNCFBaseTask(BaseTask, IOptimizationTask):  # pylint: disable=too-many-instance-attributes
    """NNCFBaseTask."""

    @check_input_parameters_type()
    def __init__(self, task_environment: TaskEnvironment, **kwargs):
        super().__init__(task_environment, **kwargs)

        # Set default model attributes.
        check_nncf_is_enabled()
        self._nncf_data_to_build = None
        self._nncf_state_dict_to_build: Dict[str, torch.Tensor] = {}
        self._nncf_preset = None
        self._optimization_methods: List[OptimizationMethod] = []
        self._precision = [ModelPrecision.FP32]

        # Extra control variables.
        self._training_work_dir = None
        self._is_training = False
        self._should_stop = False
        self._optimization_type = ModelOptimizationType.NNCF

        self._set_attributes_by_hyperparams()

        logger.info("Task initialization completed")

    def _set_attributes_by_hyperparams(self):
        quantization = self._hyperparams.nncf_optimization.enable_quantization
        pruning = self._hyperparams.nncf_optimization.enable_pruning
        if quantization and pruning:
            self._nncf_preset = "nncf_quantization_pruning"
            self._optimization_methods = [
                OptimizationMethod.QUANTIZATION,
                OptimizationMethod.FILTER_PRUNING,
            ]
            self._precision = [ModelPrecision.INT8]
            return
        if quantization and not pruning:
            self._nncf_preset = "nncf_quantization"
            self._optimization_methods = [OptimizationMethod.QUANTIZATION]
            self._precision = [ModelPrecision.INT8]
            return
        if not quantization and pruning:
            self._nncf_preset = "nncf_pruning"
            self._optimization_methods = [OptimizationMethod.FILTER_PRUNING]
            self._precision = [ModelPrecision.FP32]
            return
        raise RuntimeError("Not selected optimization algorithm")

    def _init_train_data_cfg(self, dataset: DatasetEntity):
        logger.info("init data cfg.")
        data_cfg = ConfigDict(data=ConfigDict())

        for cfg_key, subset in zip(
            ["train", "val"],
            [Subset.TRAINING, Subset.VALIDATION],
        ):
            subset = get_dataset(dataset, subset)
            if subset:
                data_cfg.data[cfg_key] = ConfigDict(
                    otx_dataset=subset,
                    labels=self._labels,
                )

        # Temparory remedy for cfg.pretty_text error
        for label in self._labels:
            label.hotkey = "a"
        return data_cfg

    def _init_nncf_cfg(self):
        base_dir = os.path.abspath(os.path.dirname(self.template_file_path))
        nncf_config_path = os.path.join(base_dir, "compression_config.json")

        with open(nncf_config_path, encoding="UTF-8") as nncf_config_file:
            common_nncf_config = json.load(nncf_config_file)

        optimization_config = compose_nncf_config(common_nncf_config, [self._nncf_preset])

        max_acc_drop = self._hyperparams.nncf_optimization.maximal_accuracy_degradation / 100
        if "accuracy_aware_training" in optimization_config["nncf_config"]:
            # Update maximal_absolute_accuracy_degradation
            (
                optimization_config["nncf_config"]["accuracy_aware_training"]["params"][
                    "maximal_absolute_accuracy_degradation"
                ]
            ) = max_acc_drop
            # Force evaluation interval
            self._recipe_cfg.evaluation.interval = 1
        else:
            logger.info("NNCF config has no accuracy_aware_training parameters")

        return ConfigDict(optimization_config)

    def _initialize_post_hook(self, options=None):
        super()._initialize_post_hook(options)
        assert self._recipe_cfg is not None and self._model_cfg is not None

        # TODO: more delicate configuration change control in MPA side

        # last batch size of 1 causes undefined behaviour for batch normalization
        # when initializing and training NNCF
        if self._data_cfg is not None:
            data_loader = self._recipe_cfg.data.get("train_dataloader", {})
            samples_per_gpu = data_loader.get("samples_per_gpu", self._recipe_cfg.data.get("samples_per_gpu"))
            otx_dataset = get_configs_by_keys(self._data_cfg.data.train, "otx_dataset")
            assert len(otx_dataset) == 1
            otx_dataset = otx_dataset[0]
            if otx_dataset is not None and len(otx_dataset) % samples_per_gpu == 1:
                data_loader["drop_last"] = True
                self._recipe_cfg.data["train_dataloader"] = data_loader

        # nncf does not suppoer FP16
        if "fp16" in self._recipe_cfg or "fp16" in self._model_cfg:
            remove_from_config(self._recipe_cfg, "fp16")
            remove_from_config(self._model_cfg, "fp16")
            logger.warning("fp16 option is not supported in NNCF. Switch to fp32.")

        # FIXME: nncf quantizer does not work with SAMoptimizer
        optimizer_config = self._recipe_cfg.optimizer_config
        if optimizer_config.get("type", "OptimizerHook") == "SAMOptimizerHook":
            optimizer_config.type = "OptimizerHook"
            logger.warning("Updateed SAMOptimizerHook to OptimizerHook as not supported.")

        # merge nncf_cfg
        nncf_cfg = self._init_nncf_cfg()
        self._recipe_cfg.merge_from_dict(nncf_cfg)

        # configure nncf
        nncf_config = self._recipe_cfg.get("nncf_config", {})
        if nncf_config.get("target_metric_name", None) is None:
            metric_name = self._recipe_cfg.evaluation.metric
            if isinstance(metric_name, list):
                metric_name = metric_name[0]
            nncf_config.target_metric_name = metric_name
            logger.info(f"'target_metric_name' not found in nncf config. Using {metric_name} as target metric")

        if is_accuracy_aware_training_set(nncf_config):
            # Prepare runner for Accuracy Aware
            self._recipe_cfg.runner = {
                "type": "AccuracyAwareRunner",
                "nncf_config": nncf_config,
            }

            # AccuracyAwareRunner needs to evaluate a model when it needs
            # unlike other runners counting on periodically evaluated score by 'EvalHook'.
            # To configure 'interval' to 'max_epoch' makes sure 'EvalHook' not to evaluate
            # during training.
            max_epoch = nncf_config.accuracy_aware_training.params.maximal_total_epochs
            self._recipe_cfg.evaluation.interval = max_epoch
            # Disable 'AdaptiveTrainSchedulingHook' as training is managed by AccuracyAwareRunner
            remove_from_configs_by_type(self._recipe_cfg.custom_hooks, "AdaptiveTrainSchedulingHook")

    @staticmethod
    def model_builder(
        config,
        *args,
        nncf_model_builder,
        model_config=None,
        data_config=None,
        is_export=False,
        return_compression_ctrl=False,
        **kwargs,
    ):
        """model_builder."""

        if model_config is not None or data_config is not None:
            config = deepcopy(config)
            if model_config is not None:
                config.merge_from_dict(model_config)
            if data_config is not None:
                config.merge_from_dict(data_config)

        compression_ctrl, model, = nncf_model_builder(
            config,
            distributed=False,
            *args,
            **kwargs,
        )

        if is_export:
            compression_ctrl.prepare_for_export()
            model.disable_dynamic_graph_building()

        if return_compression_ctrl:
            return compression_ctrl, model
        return model

    def _optimize(
        self,
        dataset: DatasetEntity,
        optimization_parameters: Optional[OptimizationParameters] = None,
    ):
        raise NotImplementedError

    def _optimize_post_hook(
        self,
        dataset: DatasetEntity,
        output_model: ModelEntity,
    ):
        pass

    @check_input_parameters_type({"dataset": DatasetParamTypeCheck})
    def optimize(
        self,
        optimization_type: OptimizationType,
        dataset: DatasetEntity,
        output_model: ModelEntity,
        optimization_parameters: Optional[OptimizationParameters] = None,
    ):
        """NNCF Optimization."""
        if optimization_type is not OptimizationType.NNCF:
            raise RuntimeError("NNCF is the only supported optimization")

        if optimization_parameters is not None:
            update_progress_callback = optimization_parameters.update_progress
        else:
            update_progress_callback = default_progress_callback

        self._time_monitor = OptimizationProgressCallback(
            update_progress_callback,
            loading_stage_progress_percentage=5,
            initialization_stage_progress_percentage=5,
        )

        self._data_cfg = self._init_train_data_cfg(dataset)
        self._is_training = True

        results = self._optimize(dataset, optimization_parameters)

        # Check for stop signal when training has stopped.
        # If should_stop is true, training was cancelled
        if self._should_stop:
            logger.info("Training cancelled.")
            self._should_stop = False
            self._is_training = False
            return

        model_ckpt = results.get("final_ckpt")
        if model_ckpt is None:
            logger.error("cannot find final checkpoint from the results.")
            # output_model.model_status = ModelStatus.FAILED
            return
        # update checkpoint to the newly trained model
        self._model_ckpt = model_ckpt

        self._optimize_post_hook(dataset, output_model)

        self.save_model(output_model)

        output_model.model_format = ModelFormat.BASE_FRAMEWORK
        output_model.optimization_type = self._optimization_type
        output_model.optimization_methods = self._optimization_methods
        output_model.precision = self._precision

        self._is_training = False

    def _save_model_post_hook(self, modelinfo):
        pass

    @check_input_parameters_type()
    def save_model(self, output_model: ModelEntity):
        """Saving model function for NNCF Task."""
        assert self._recipe_cfg is not None
        assert self._model_cfg is not None

        buffer = io.BytesIO()
        hyperparams_str = ids_to_strings(cfg_helper.convert(self._hyperparams, dict, enum_to_str=True))
        labels = {label.name: label.color.rgb_tuple for label in self._labels}

        config = deepcopy(self._recipe_cfg)
        config.merge_from_dict(self._model_cfg)

        def update(d, u):  # pylint: disable=invalid-name
            for k, v in u.items():  # pylint: disable=invalid-name
                if isinstance(v, Mapping):
                    d[k] = update(d.get(k, {}), v)
                else:
                    d[k] = v
            return d

        modelinfo = torch.load(self._model_ckpt, map_location=torch.device("cpu"))
        modelinfo = update(
            dict(model=modelinfo),
            {
                "meta": {
                    "nncf_enable_compression": True,
                    "config": config,
                },
                "config": hyperparams_str,
                "labels": labels,
                "VERSION": 1,
            },
        )
        self._save_model_post_hook(modelinfo)

        torch.save(modelinfo, buffer)
        output_model.set_data("weights.pth", buffer.getvalue())
        output_model.set_data(
            "label_schema.json",
            label_schema_to_bytes(self._task_environment.label_schema),
        )



"""Model deployment tool."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import os
from pathlib import Path

from otx.api.configuration.helper import create
from otx.api.entities.model import ModelEntity
from otx.api.entities.task_environment import TaskEnvironment
from otx.cli.manager import ConfigManager
from otx.cli.utils.importing import get_impl_class
from otx.cli.utils.io import read_label_schema, read_model
from otx.cli.utils.parser import get_parser_and_hprams_data


def get_args():
    """Parses command line arguments."""
    parser, _, _ = get_parser_and_hprams_data()

    parser.add_argument(
        "--load-weights",
        help="Load model weights from previously saved checkpoint.",
    )
    parser.add_argument(
        "--save-model-to",
        help="Location where openvino.zip will be stored.",
    )

    return parser.parse_args()


def main():
    """Main function that is used for model evaluation."""

    # Parses input arguments.
    args = get_args()
    config_manager = ConfigManager(args, mode="deploy")
    # Auto-Configuration for model template
    config_manager.configure_template()

    # Reads model template file.
    template = config_manager.template

    # Get hyper parameters schema.
    hyper_parameters = template.hyper_parameters.data
    assert hyper_parameters

    if not args.load_weights and config_manager.check_workspace():
        exported_weight_path = config_manager.workspace_root / "models-exported/openvino.xml"
        if not exported_weight_path.exists():
            raise RuntimeError("OpenVINO-exported models are supported.")
        args.load_weights = str(exported_weight_path)

    # Get classes for Task, ConfigurableParameters and Dataset.
    if not args.load_weights.endswith(".bin") and not args.load_weights.endswith(".xml"):
        raise RuntimeError("Only OpenVINO-exported models are supported.")

    task_class = get_impl_class(template.entrypoints.openvino)

    environment = TaskEnvironment(
        model=None,
        hyper_parameters=create(hyper_parameters),
        label_schema=read_label_schema(args.load_weights),
        model_template=template,
    )
    environment.model = read_model(environment.get_model_configuration(), args.load_weights, None)

    task = task_class(task_environment=environment)

    deployed_model = ModelEntity(None, environment.get_model_configuration())

    if "save_model_to" not in args or not args.save_model_to:
        args.save_model_to = str(config_manager.workspace_root / "model-deployed")
    os.makedirs(args.save_model_to, exist_ok=True)
    task.deploy(deployed_model)
    with open(Path(args.save_model_to) / "openvino.zip", "wb") as write_file:
        write_file.write(deployed_model.exportable_code)

    return dict(retcode=0, template=template.name)


if __name__ == "__main__":
    main()



"""setup file for demo package."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from pathlib import Path

from setuptools import find_packages, setup

SETUP_DIR = Path(__file__).resolve().parent

with open(SETUP_DIR / "requirements.txt", "r", encoding="utf8") as f:
    required = f.read().splitlines()

packages = find_packages(str(SETUP_DIR))
package_dir = {packages[0]: str(SETUP_DIR / packages[0])}

setup(
    name=packages[0],
    version="0.0",
    author="Intel® Corporation",
    license="Copyright (c) 2021-2022 Intel Corporation. " "SPDX-License-Identifier: Apache-2.0",
    description="Demo based on ModelAPI classes",
    packages=packages,
    package_dir=package_dir,
    package_data={
        packages[0]: ["*.json"],
    },
    install_requires=required,
)



"""Model configuration of Fast RCNN with X3D for Action Detection Task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=invalid-name

# model setting
model = dict(
    type="AVAFastRCNN",
    backbone=dict(type="X3D", gamma_w=1, gamma_b=2.25, gamma_d=2.2),
    roi_head=dict(
        type="AVARoIHead",
        bbox_roi_extractor=dict(
            type="SingleRoIExtractor3D", roi_layer_type="RoIAlign", output_size=8, with_temporal_pool=True
        ),
        bbox_head=dict(type="BBoxHeadAVA", in_channels=432, num_classes=81, multilabel=False, dropout_ratio=0.5),
    ),
    train_cfg=dict(
        rcnn=dict(
            assigner=dict(type="MaxIoUAssignerAVA", pos_iou_thr=0.9, neg_iou_thr=0.9, min_pos_iou=0.9),
            sampler=dict(type="RandomSampler", num=32, pos_fraction=1, neg_pos_ub=-1, add_gt_as_proposals=True),
            pos_weight=1.0,
            debug=False,
        )
    ),
    test_cfg=dict(rcnn=dict(action_thr=0.002)),
)

optimizer = dict(type="SGD", lr=0.1, momentum=0.9, weight_decay=1e-5)
optimizer_config = dict(grad_clip=dict(max_norm=40, norm_type=2))

lr_config = dict(
    policy="CosineAnnealing",
    by_epoch=False,
    min_lr=0,
    warmup="linear",
    warmup_by_epoch=True,
    warmup_iters=2,
    warmup_ratio=0.1,
)
checkpoint_config = dict(interval=1)
workflow = [("train", 1)]
evaluation = dict(interval=1, save_best="mAP@0.5IOU", final_metric="mAP@0.5IOU")
log_config = dict(
    interval=10,
    hooks=[
        dict(type="TextLoggerHook"),
    ],
)
dist_params = dict(backend="nccl")
log_level = "INFO"
work_dir = "logs/x3d_kinetics_pretrained_ava_rgb/cosine/"
load_from = (
    "https://download.openmmlab.com/mmaction/recognition/x3d/facebook/"
    "x3d_m_facebook_16x5x1_kinetics400_rgb_20201027-3f42382a.pth"
)
resume_from = None
find_unused_parameters = False
# Temporary solution, gpu_ids is not used in otx
gpu_ids = [0]
seed = 2
